#!/usr/bin/env python
# coding: utf-8

import copy
import logging
import threading
import time

from kazoo import exceptions
from kazoo.client import KazooState
from kazoo.exceptions import KazooException
from kazoo.exceptions import NoNodeError

from pykit import rangeset
from pykit import utfjson
from pykit import zkutil

from .exceptions import ConnectionLoss
from .exceptions import Deadlock
from .exceptions import HigherTXApplied
from .exceptions import NotLocked
from .exceptions import RetriableError
from .exceptions import TXTimeout
from .exceptions import UnlockNotAllowed
from .exceptions import UserAborted
from .status import ABORTED
from .status import COMMITTED
from .status import PAUSED
from .status import UNKNOWN
from .zkstorage import ZKStorage

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 10


class TXRecord(object):

    def __init__(self, k, v, txid, version, values):
        self.k = k
        self.v = v             # the latest updated value
        self.txid = txid       # the txid in which the `v` is written in `k`
        self.version = version  # stat.version of zk node `k`
        self.values = values

    def __str__(self):
        return '{k}={v}@{txid}'.format(
            k=self.k, v=self.v, txid=txidstr(self.txid))


class ZKTransaction(object):

    def __init__(self, zk, txid=None, timeout=None):

        if timeout is None:
            timeout = DEFAULT_TIMEOUT

        # Save the original arg for self.run()
        self._zk = zk

        self.zke, self.owning_zk = zkutil.kazoo_client_ext(zk)
        self.timeout = timeout

        self.txid = txid
        self.expire_at = None
        self.tx_alive_lock = None

        # User locked and required keys
        self.got_keys = {}
        # Keys user need to commit
        self.modifications = {}

        self.state_lock = threading.RLock()
        self.connected = True

        self.tx_status = None

        self.zkstorage = ZKStorage(self.zke)

        self.zke.add_listener(self._on_conn_change)

    def _on_conn_change(self, state):

        logger.debug('state changed: {state}'.format(state=state,))

        with self.state_lock:
            if state == KazooState.LOST or state == KazooState.SUSPENDED:
                self.connected = False

    def lock_get(self, key, blocking=True, latest=True):

        # We use persistent lock(ephemeral=False)
        # thus we do not need to care about connection loss during locking
        # phase.
        # But we still need to check connection when creating journal.

        if key in self.got_keys:
            if latest and key in self.modifications:
                return copy.deepcopy(self.modifications[key])

            return copy.deepcopy(self.got_keys[key])

        if blocking:
            self.lock_key(key)
        else:
            locked, other_txid, ver = self.try_lock_key(key)
            if not locked:
                return None

        val, version = self.zkstorage.record.get(key)
        ltxid, lvalue = val[-1]

        curr = TXRecord(k=key, v=copy.deepcopy(lvalue), txid=ltxid,
                        version=version, values=val)
        self.got_keys[key] = curr

        if ltxid > self.txid:
            raise HigherTXApplied('{tx} seen a higher txid applied: {txid}'.format(
                tx=self, txid=ltxid))

        rec = TXRecord(k=key, v=copy.deepcopy(lvalue), txid=ltxid,
                       version=version, values=val)
        return rec

    def unlock(self, rec):

        if rec.k not in self.got_keys:
            raise NotLocked("Not allowed to unlock non-locked: {k}".format(k=rec.k))

        if rec.k in self.modifications:
            raise UnlockNotAllowed('{k} has set as changed'.format(k=rec.k))

        self.zkstorage.try_release_key(self.txid, rec.k)

        del self.got_keys[rec.k]

    def set(self, rec):
        logger.info('{tx} tx.set: {rec}'.format(tx=self, rec=rec))
        self.modifications[rec.k] = rec

    def set_state(self, state_data):
        txst = {
            'got_keys': self.got_keys.keys(),
            'data': state_data,
        }

        self.zkstorage.state.set_or_create(self.txid, txst)

    def get_state(self, txid=None):

        txst, ver = self._get_state(txid=txid)
        if txst is None:
            return None
        else:
            return txst['data']

    def _get_state(self, txid=None):
        if txid is None:
            txid = self.txid
        txst, ver = self.zkstorage.state.get(txid)

        if txst is None:
            return None, -1

        return txst, ver

    def has_state(self, txid):
        st, _ = self.zkstorage.state.get(txid)
        return st is not None

    def delete_state(self, txid):
        try:
            self.zkstorage.state.delete(txid)
        except NoNodeError:
            pass

    def lock_key(self, key):

        try:
            self._lock_key(key)
        except zkutil.LockTimeout:
            raise TXTimeout('{tx} timeout waiting for lock: {key}'.format(tx=self, key=key))

        logger.info('{tx} [{key}] locked'.format(tx=self, key=key))

    def try_lock_key(self, key):

        for other_txid, ver in self.zkstorage.acquire_key_loop(
                self.txid, key, timeout=-1):

            return False, other_txid, ver

        return True, self.txid, -1

    def _lock_key(self, key):

        for other_txid, ver in self.zkstorage.acquire_key_loop(
                self.txid, key, timeout=self.time_left()):

            logger.info('{tx} wait[{key}]-> {other_txid} ver: {ver}'.format(
                tx=self, key=key, other_txid=txidstr(other_txid), ver=ver))

            # A tx that has state saved is recoverable, is treated as alive
            # tx.
            if not self.is_tx_alive(other_txid) and not self.has_state(other_txid):

                rst = self.redo_dead_tx(other_txid, timeout=self.expire_at-time.time())

                # aborted tx did not release lock by itself
                if rst == ABORTED:
                    self.zkstorage.try_release_key(other_txid, key)

                continue

            logger.info('{tx}: other tx {other_txid} is alive'.format(
                tx=self, other_txid=txidstr(other_txid)))

            if self.txid > other_txid:

                if len(self.got_keys) == 0:

                    # no locking, no deadlock
                    logger.info('{tx} wait[{key}]-> {other_txid} no-lock'.format(
                        tx=self, key=key, other_txid=txidstr(other_txid)))

                    for i in range(other_txid, self.txid):
                        self.wait_tx_to_finish(i)
                    continue

                else:

                    # let earlier tx to finish first
                    logger.info('{tx} wait[{key}]-> {other_txid} deadlock'.format(
                        tx=self, key=key, other_txid=txidstr(other_txid)))

                    self.release_all_key_locks()
                    self.delete_state(self.txid)

                    for i in range(other_txid, self.txid):
                        self.wait_tx_to_finish(i)

                    raise Deadlock('my txid: {mytxid} lockholder txid: {other_txid}'.format(
                        mytxid=self.txid, other_txid=other_txid))

            # self.txid < other_txid:
            # I am an earlier tx. I should run first.
            #
            # If there is no deadlock:
            #   the older tx will finish later. Then I wait.
            #
            # If there is a deadlock:
            #   the older tx will abort.
            logger.info('{tx} wait[{key}]-> {other_txid}'.format(
                tx=self, key=key, other_txid=txidstr(other_txid)))

    def wait_tx_to_finish(self, txid):

        logger.info('{tx} wait-> {other_txid}'.format(tx=self, other_txid=txidstr(txid)))

        try:
            zkutil.wait_absent(self.zke,
                               self.zke._zkconf.tx_alive(txid),
                               timeout=self.time_left())
        except zkutil.ZKWaitTimeout as e:

            logger.info(repr(e) + ' while waiting for other tx: {txid}'.format(
                txid=txid))

            raise TXTimeout('{tx} timeout waiting for tx:{txid}'.format(tx=self, txid=txid))

    def redo_all_dead_tx(self):

        sets, ver = self.zkstorage.txidset.get()

        # Only check for holes in the txid range set.
        # With `txidset = [[1, 2], [4, 5]]`, tx-2 and tx-3 is unfinished tx.
        # Check these tx

        # A tx is committed or aborted that does not need to handle it again.
        known = rangeset.union(sets[COMMITTED], sets[ABORTED])

        for rng in known[:-1]:

            txid = rng[1]

            if not self.is_tx_alive(txid) and not self.has_state(txid):
                self.redo_dead_tx(txid)

    def redo_dead_tx(self, txid, timeout=10):

        expire_at = time.time() + timeout
        dead_tx = ZKTransaction(self.zke,
                                timeout=expire_at - time.time())

        dead_tx.begin(txid=txid)
        rst = dead_tx.redo()
        dead_tx._close()
        return rst

    def redo(self):

        logger.info('REDO: {tx}'.format(tx=self))

        txidset, ver = self.zkstorage.txidset.get()
        logger.info('{tx} got txidset: {txidset}'.format(tx=self, txidset=txidset))

        if txidset[COMMITTED].has(self.txid):
            return COMMITTED
        elif txidset[ABORTED].has(self.txid):
            return ABORTED

        # self.txid is not in txidset

        try:
            jour, version = self.zkstorage.journal.get(self.txid)
        except NoNodeError:
            # This tx has not yet written a journal.
            # Nothing to apply
            self.zkstorage.add_to_txidset(ABORTED, self.txid)
            return ABORTED

        self.zkstorage.add_to_txidset(COMMITTED, self.txid)
        return COMMITTED

    def is_tx_alive(self, txid):
        try:
            self.zke.get(self.zke._zkconf.tx_alive(txid))
            logger.info('{tx} tx:{txid} is alive'.format(tx=self, txid=txid))
            return True
        except NoNodeError:
            logger.info('{tx} tx:{txid} is dead'.format(tx=self, txid=txid))
            return False

    def begin(self, txid=None):

        assert self.expire_at is None

        self.expire_at = time.time() + self.timeout

        if self.txid is None:

            if txid is not None:
                # Run a specified tx, normally when to recover a dead tx process
                self.txid = txid

            else:
                # Run a new tx, create txid.
                zstat = self.zke.set(self.zke._zkconf.txid_maker(), 'x')
                self.txid = zstat.version

        zkconf = copy.deepcopy(self.zke._zkconf.conf)
        zkconf['lock_dir'] = self.zke._zkconf.tx_alive()

        self.tx_alive_lock = zkutil.ZKLock(self.txid,
                                           zkconf=zkconf,
                                           zkclient=self.zke,
                                           timeout=self.expire_at-time.time())

        try:
            self.tx_alive_lock.acquire()
        except zkutil.LockTimeout as e:
            raise TXTimeout(repr(e) + ' while waiting for tx alive lock')

    def commit(self):

        # Only when commit, it is necessary to ensure connection still active:
        # Thus tx_alive_lock is not lost, then no other process would take
        # charge of this tx.

        if self.time_left() < 0:
            raise TXTimeout('{tx} timeout when committing'.format(tx=self))

        self._assert_connected()

        kazootx = self.zke.transaction()

        jour = {}
        cnf = self.zke._zkconf

        for k, rec in self.modifications.items():
            curr = self.got_keys[k]
            if rec.v == curr.v:
                continue

            jour[k] = rec.v
            if curr.version == -1:
                kazootx.create(cnf.record(rec.k),
                               utfjson.dump(curr.values + [[self.txid, rec.v]]))
            else:
                kazootx.set_data(cnf.record(rec.k),
                                 utfjson.dump(curr.values + [[self.txid, rec.v]]),
                                 version=curr.version)

        kazootx.create(cnf.journal(self.txid),
                       utfjson.dump(jour))

        state, ver = self._get_state(self.txid)
        if ver > -1:
            kazootx.delete(cnf.tx_state(self.txid), version=ver)

        for key in self.got_keys:
            kazootx.delete(cnf.lock(key), version=0)

        kazootx.commit()

        self.zkstorage.add_to_txidset(COMMITTED, self.txid)

        logger.info('{tx} updated txidset: {status}'.format(
            tx=self, status=COMMITTED))

        self.tx_status = COMMITTED
        self.modifications = {}
        self._close()

    def abort(self):
        raise UserAborted()

    def is_timeout(self):
        return self.time_left() > 0

    def time_left(self):
        return self.expire_at - time.time()

    def _close(self):

        self.release_all_key_locks()

        if self.tx_alive_lock is not None:

            try:
                self.tx_alive_lock.release()

            except (exceptions.ConnectionClosedError,
                    exceptions.ConnectionLoss) as e:

                logger.warn('{tx} {e} while release tx alive lock'.format(
                    tx=self, e=repr(e)))

            self.tx_alive_lock = None

        self.zke.remove_listener(self._on_conn_change)

        self.modifications = {}

        if self.owning_zk:
            try:
                self.zke.stop()
            except KazooException as e:
                logger.info(repr(e) + ' while zkclient.stop()')

    def release_all_key_locks(self):

        for key in self.got_keys:

            logger.info('{tx} releasing:[{key}]'.format(tx=self, key=key))

            try:
                self.zkstorage.try_release_key(self.txid, key)
            except (exceptions.ConnectionClosedError,
                    exceptions.ConnectionLoss) as e:

                logger.warn('{tx} {e} while release key lock {key}'.format(
                    tx=self, e=repr(e), key=key))
                break

        self.got_keys = {}

    def _assert_connected(self):
        with self.state_lock:
            if not self.connected:
                raise ConnectionLoss()

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, errtype, errval, _traceback):

        logger.info('{tx} __exit__ errtype:{errtype} errval:{errval}'.format(
            tx=self, errtype=errtype, errval=errval))

        try:
            if self.tx_status is None:

                try:
                    has = self.has_state(self.txid)
                    if has:
                        self.tx_status = PAUSED

                    else:
                        self.tx_status = ABORTED
                        self.zkstorage.add_to_txidset(ABORTED, self.txid)
                except (exceptions.ConnectionClosedError,
                        exceptions.ConnectionLoss) as e:

                    logger.warn('{tx} {e} while adding to txidset.ABORTED'.format(
                        tx=self, e=repr(e)))

                if self.tx_status is None:
                    self.tx_status = UNKNOWN

            if errval is None:
                return True

            if isinstance(errval, UserAborted):
                if self.txid is not None:
                    self.delete_state(self.txid)
                return True

            return False

        finally:
            self._close()

    def __str__(self):
        return '{tx}[{locked}]'.format(tx=txidstr(self.txid),
                                       locked=','.join(sorted(self.got_keys.keys())))


def run_tx(zk, func, timeout=None, args=(), kwargs=None):

    if timeout is None:
        timeout = 10

    if kwargs is None:
        kwargs = {}

    expire_at = time.time() + timeout

    while True:

        tx = ZKTransaction(zk, timeout=expire_at - time.time())

        try:
            with tx:
                func(tx, *args, **kwargs)

            assert tx.tx_status is not None, 'with-statement guarantees tx_status is not None'
            return tx.tx_status, tx.txid

        except RetriableError as e:
            logger.info('{tx} {e}'.format(tx=tx, e=repr(e)))
            continue


def list_recoverable(zk):

    t = ZKTransaction(zk)

    try:
        path = t.zkstorage.state.get_path('')
        children = t.zke.get_children(path)

        for txid in children:
            txid = int(txid)

            st = t.get_state(txid)
            if st is not None and not t.is_tx_alive(txid):
                yield txid, st

    finally:
        t._close()


def txidstr(txid):
    return 'tx-{txid:0>10}'.format(txid=txid)



#!/usr/bin/env python2
# coding: utf-8


import errno
import logging
import os
import stat
import time

import __main__
from pykit import config
from pykit import portlock
from pykit import utfjson

from . import fsutil

logger = logging.getLogger(__name__)


no_data_timeout = 3600
read_size = 1024 * 1024 * 16
file_check_time_range = (0.05, 1.0)  # sec
stat_dir = config.cat_stat_dir or '/tmp'
SEEK_START = 'start'
SEEK_END = 'end'


class CatError(Exception):
    pass


class NoData(CatError):
    pass


class NoSuchFile(CatError):
    pass


class LockTimeout(CatError):
    pass


class FakeLock(object):

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        return


class Cat(object):

    def __init__(self, fn,
                 handler=None,
                 file_end_handler=None,
                 exclusive=True,
                 id=None,
                 strip=False,
                 read_chunk_size=read_size):

        self.fn = fn
        self.handler = handler
        self.file_end_handler = file_end_handler
        self.exclusive = exclusive

        if id is None:

            if hasattr(__main__, '__file__'):
                id = __main__.__file__
                id = os.path.basename(id)
                if id == '<stdin>':
                    id = '__stdin__'
            else:
                id = '__instant_command__'

        self.id = id
        self.strip = strip

        read_chunk_size = int(read_chunk_size)
        if read_chunk_size <= 0:
            read_chunk_size = read_size
        self.read_chunk_size = read_chunk_size

        self.running = None
        self.bufferred = None  # (offset, content)

        if (self.handler is not None
                and callable(self.handler)):

            self.handler = [self.handler]

    def cat(self, timeout=None, default_seek=None):

        for line in self.iterate(timeout=timeout, default_seek=default_seek):

            for h in self.handler:
                try:
                    h(line)
                except Exception as e:
                    logger.exception(repr(e)
                                     + ' while handling {line}'.format(line=repr(line)))

    def iterate(self, timeout=None, default_seek=None):
        self.running = True
        try:
            for x in self._iter(timeout, default_seek):
                yield x
        finally:
            self.running = False

    def _iter(self, timeout, default_seek):

        if timeout == 0:
            # timeout at once after one read, if there is no more data in file.
            # set to -1 to prevent possible slight time shift.
            timeout = -1

        if self.exclusive:
            lck = portlock.Portlock(self.lock_name(), timeout=0.1)
        else:
            lck = FakeLock()

        try:
            with lck:
                for x in self._nolock_iter(timeout, default_seek):
                    yield x
        except portlock.PortlockTimeout:
            raise LockTimeout(self.id, self.fn, 'other Cat() has been holding this lock')

    def _nolock_iter(self, timeout, default_seek):

        if timeout is None:
            timeout = no_data_timeout

        expire_at = time.time() + timeout

        while True:

            # NOTE: Opening a file and waiting for new data in it does not work.
            #
            # It has to check for file overriding on fs periodically.
            # Or it may happens that it keeps waiting for new data on a deleted
            # but opened file, while new data is actually written into a new
            # file with the same path.
            #
            # Thus we check if file changed for about 5 times(by reopening it).

            read_timeout = (expire_at - time.time()) / 5.0

            if read_timeout < file_check_time_range[0]:
                read_timeout = file_check_time_range[0]

            if read_timeout > file_check_time_range[1]:
                read_timeout = file_check_time_range[1]

            f = self.wait_open_file(timeout=expire_at - time.time())

            with f:
                try:
                    for x in self.iter_to_file_end(
                            f, read_timeout, default_seek):
                        yield x

                    # re-new expire_at if there is any data read.
                    expire_at = time.time() + timeout
                    if time.time() > expire_at:
                        # caller expect to return at once when it has read once,
                        # timeout < 0

                        # When timeout, it means no more data will be appended.
                        # Thus the bufferred must be a whole line, even there is
                        # not a trailing '\n' presents
                        if self.bufferred is not None:
                            l = self.bufferred[1]
                            self.bufferred = None
                            yield l
                        return

                except NoData as e:

                    # NoData raises only when there is no data yield.

                    logger.info(repr(e) + ' while cat: {fn}'.format(fn=self.fn))

                    if time.time() > expire_at:
                        # raise last NoData
                        raise

    def wait_open_file(self, timeout):

        expire_at = time.time() + timeout
        sleep_time = 0.01
        max_sleep_time = 1

        f = self._try_open_file()
        if f is not None:
            return f

        logger.info('file not found: {fn}'.format(fn=self.fn))

        while time.time() < expire_at:

            f = self._try_open_file()
            if f is not None:
                return f

            sl = min([sleep_time, expire_at - time.time()])
            logger.debug('file not found: {fn}, sleep for {sl}'.format(fn=self.fn, sl=sl))
            time.sleep(sl)

            sleep_time = min([sleep_time * 1.5, max_sleep_time])

        logger.warn('file not found'
                    ' while waiting for it to be present: {fn}'.format(fn=self.fn))

        raise NoSuchFile(self.fn)

    def iter_to_file_end(self, f, read_timeout, default_seek):

        full_chunk_timeout = min([read_timeout * 0.01, 1])

        offset = self.get_last_offset(f, default_seek)
        f.seek(offset)

        logger.info('scan {fn} from offset: {offset}'.format(
            fn=self.fn,
            offset=offset))

        self.wait_for_new_data(f, full_chunk_timeout, read_timeout)
        for line in self.iter_lines(f):
            logger.debug('yield:' + repr(line))
            yield line

        if self.file_end_handler is not None:
            self.file_end_handler()

    def wait_for_new_data(self, f, full_chunk_timeout, timeout):

        # Before full_chunk_expire_at, wait for a full chunk data to be ready to
        # maximize throughput.
        # If time exceeds full_chunk_expire_at, return True if there is any data ready.

        full_chunk_expire_at = time.time() + full_chunk_timeout
        expire_at = time.time() + timeout

        while True:

            if f.tell() + self.read_chunk_size < _file_size(f):
                return

            if (f.tell() < _file_size(f)
                    and time.time() > full_chunk_expire_at):
                return

            if time.time() >= expire_at:
                raise NoData()

            time.sleep(0.05)

    def stat_path(self):
        return os.path.join(stat_dir, self.lock_name())

    def lock_name(self):
        name = os.path.realpath(self.fn)
        name = 'fsutil_cat_lock_!' + self.id + '!'.join(name.split('/'))
        return name

    def read_last_stat(self):

        cont = fsutil.read_file(self.stat_path())
        if cont.startswith('{'):
            return utfjson.load(cont)

        # old format: TODO remove it

        last = cont.strip().split(' ')
        if len(last) != 3:
            raise IOError('InvalidRecordFormat', last)

        (lastino, lastsize, lastoff) = last

        lastino = int(lastino)
        lastoff = int(lastoff)

        return {
            "inode": lastino,
            "offset": lastoff,
        }

    def write_last_stat(self, f, offset):

        st = os.fstat(f.fileno())

        ino = st[stat.ST_INO]

        last = {
            "inode": ino,
            "offset": offset,
        }

        fsutil.write_file(self.stat_path(), utfjson.dump(last))

        logger.info('position written fn=%s inode=%d offset=%d' % (
            self.fn, ino, offset))

    def get_last_offset(self, f, default_seek):
        st = os.fstat(f.fileno())
        ino = st[stat.ST_INO]
        size = st[stat.ST_SIZE]

        max_residual = None

        if default_seek is None or default_seek == SEEK_START:
            default_offset = 0

        elif default_seek == SEEK_END:
            default_offset = size

        elif default_seek < 0:
            max_residual = 0 - default_seek
            default_offset = max(size - max_residual, 0)

        else:
            default_offset = default_seek

        stat_file = self.stat_path()

        if not os.path.isfile(stat_file):
            return default_offset

        try:
            last = self.read_last_stat()
        except (IOError, ValueError):
            # damaged stat file
            return default_offset

        if max_residual is not None:
            if size - last['offset'] > max_residual:
                last['offset'] = size - max_residual

        if last['inode'] != ino or last['offset'] > size:
            return default_offset

        return last['offset']

    def iter_lines(self, f):

        while True:
            offset = f.tell()
            fsize = _file_size(f)
            if offset >= fsize:
                break

            lines = f.readlines(self.read_chunk_size)
            if self.bufferred is not None:
                offset = self.bufferred[0]
                lines[0] = self.bufferred[1] + lines[0]
                self.bufferred = None

            try:
                for l in lines[:-1]:
                    line = l
                    if self.strip:
                        line = line.strip('\r\n')

                    offset += len(l)
                    yield line

                l = lines[-1]
                if not l.endswith(('\r', '\n')):
                    self.bufferred = (offset, l)
                    offset += len(l)
                else:
                    line = l
                    if self.strip:
                        line = line.strip('\r\n')
                    offset += len(l)
                    yield line
            finally:
                self.write_last_stat(f, offset)

    def _try_open_file(self):
        try:
            f = open(self.fn)
            logger.info('file found and opened {fn}'.format(fn=self.fn))
            return f
        except IOError as e:
            if e.errno == errno.ENOENT:
                pass

        return None

    def reset_stat(self):
        stat_path = self.stat_path()
        if not os.path.isfile(stat_path):
            return

        os.remove(stat_path)


def _file_size(f):
    st = os.fstat(f.fileno())
    return st[stat.ST_SIZE]



"""StopWatch - library for adding timers and tags in your code for performance monitoring
https://github.com/dropbox/stopwatch

StopWatch operates on a notion of "spans" which represent scopes of code for which we
want to measure timing. Spans can be nested and placed inside loops for aggregation.

For example:
```
with sw.timer('root'):
    for i in range(50):
        with sw.timer('inner_task'):
            do_inner_task(i)
```

StopWatch requires a root scope which upon completion signifies the end of the round
of measurements. On a server, you might use a single request as your root scope.

StopWatch produces two kinds of reports.
1) Aggregated (see _reported_values).
2) Non-aggregated or "tracing" (see _reported_traces)

Aggregated reports have one value for each timer name. Great for averaging across
various root scopes / requests. The above example would have 2 reports.

Tracing reports keep one trace every time we enter a scope. The above example would
have 51 reports. This is great for digging into a data for a particular request,
or constructing waterfall graphs.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import contextlib
import random as insecure_random
import time

TraceAnnotation = collections.namedtuple('TraceKeyValueAnnotation', ['key', 'value', 'time'])
AggregatedReport = collections.namedtuple('AggregatedReport',
                                          ['aggregated_values', 'root_timer_data'])

class TimerData(object):
    """
    Simple object that wraps all data needed for a single timer span.
    The StopWatch object maintains a stack of these timers.
    """

    __slots__ = (
        'span_id',
        'name',
        'start_time',
        'end_time',
        'trace_annotations',
        'parent_span_id',
        'log_name',
    )

    def __init__(self, name, start_time, parent_name):
        # Generate new span id.
        self.span_id = '%032x' % insecure_random.getrandbits(128)
        self.name = name
        self.start_time = start_time
        self.end_time = None  # Gets filled in later
        self.trace_annotations = []
        self.parent_span_id = None  # Gets filled in at the end

        if parent_name:
            self.log_name = parent_name + '#' + name
        else:
            self.log_name = name

    def __repr__(self):
        return ('name=%r, span_id=%r start_time=%r end_time=%r annotations=%r, parent_span_id=%r,'
                'log_name=%r') % (
            self.name,
            self.span_id,
            self.start_time,
            self.end_time,
            self.trace_annotations,
            self.parent_span_id,
            self.log_name,
        )

def format_report(aggregated_report):
    """returns a pretty printed string of reported values"""
    values = aggregated_report.aggregated_values
    root_tr_data = aggregated_report.root_timer_data

    # fetch all values only for main stopwatch, ignore all the tags
    log_names = sorted(
        log_name for log_name in values if "+" not in log_name
    )
    if not log_names:
        return

    root = log_names[0]
    root_time_ms, root_count, bucket = values[root]
    buf = [
        "************************",
        "*** StopWatch Report ***",
        "************************",
        "%s    %.3fms (%.f%%)" % (root.ljust(20), root_time_ms / root_count, 100),
    ]
    for log_name in log_names[1:]:
        delta_ms, count, bucket = values[log_name]
        depth = log_name[len(root):].count("#")
        short_name = log_name[log_name.rfind("#") + 1:]
        bucket_name = bucket.name if bucket else ""

        buf.append("%s%s    %s %4d  %.3fms (%.f%%)" % (
            "    " * depth, bucket_name.ljust(12),
            short_name.ljust(20),
            count,
            delta_ms,
            delta_ms / root_time_ms * 100.0,
        ))

    annotations = sorted(ann.key for ann in root_tr_data.trace_annotations)
    buf.append("Annotations: %s" % (', '.join(annotations)))
    return "\n".join(buf)

def default_export_tracing(reported_traces):
    """Default implementation of non-aggregated trace logging"""
    pass

def default_export_aggregated_timers(aggregated_report):
    """Default implementation of aggregated timer logging"""
    pass

class StopWatch(object):
    """StopWatch - main class for storing timer stack and exposing timer functions/contextmanagers
    to the rest of the code"""

    def __init__(self,
                 strict_assert=True,
                 export_tracing_func=None,
                 export_aggregated_timers_func=None,
                 max_tracing_spans_for_path=1000,
                 min_tracing_milliseconds=3,
                 time_func=None):
        """
        Arguments:
          strict_assert: If True, assert on callsite misuse

          export_tracing_func: Function to log tracing data when stack empties

          export_aggregated_timers_func: Function to log timers when stack empties

          max_tracing_spans_for_path:
            The maximum number of spans to be logged per root scope for
            each unique path, so we make sure we aren't too excessive on the
            non-aggregated tracing report

          min_tracing_milliseconds:
            To reduce the large number of trivial spans, don't trace
            anything that takes less then this amount of time.

          time_func:
            Function which returns the current time in seconds. Defaults to time.time
        """

        self._timer_stack = []
        self._strict_assert = strict_assert
        self._export_tracing_func = export_tracing_func or default_export_tracing
        self._export_aggregated_timers_func = (
            export_aggregated_timers_func or default_export_aggregated_timers
        )
        self._time_func = time_func or time.time
        self.MAX_REQUEST_TRACING_SPANS_FOR_PATH = max_tracing_spans_for_path
        self.TRACING_MIN_NUM_MILLISECONDS = min_tracing_milliseconds
        self._last_trace_report = None
        self._last_aggregated_report = None

        self._reset()

    def _reset(self):
        """Reset internal timer stack when stack is cleared"""
        if self._timer_stack:
            assert not self._strict_assert, \
                "StopWatch reset() but stack not empty: %r" % (self._timer_stack,)
        self._reported_values = {}
        self._reported_traces = []
        self._root_annotations = []
        self._slow_annotations = {}

        # Dictionary of span names that have been cancelled in the current
        # context. Used to ensure that a cancelled span is not redundantly ended as well.
        self._cancelled_spans = {}

    ################
    # Public methods
    ################
    @contextlib.contextmanager
    def sampling_timer(self, name, p, *n, **kwargs):
        """Context manager that will time the context with probability p."""
        if p > insecure_random.uniform(0.0, 1.0):
            with self.timer(name, *n, **kwargs):
                yield
        else:
            yield

    @contextlib.contextmanager
    def timer(self, name, bucket=None, start_time=None, end_time=None):
        """Context manager to wrap a stopwatch span"""
        self.start(name, start_time=start_time)
        try:
            yield
        except Exception as e:
            self.add_annotation('Exception', type(e).__name__, event_time=end_time)
            raise
        finally:
            if name in self._cancelled_spans:
                del self._cancelled_spans[name]
            else:
                self.end(name, end_time=end_time, bucket=bucket)

    def start(self, name, start_time=None):
        """Begin a stopwatch span
        Arguments:
            name:
                Name of the span to start
            start_time:
                Time (s) at which the scope began if set. (if not, use the current time)
        """
        if start_time is None:
            start_time = self._time_func()
        self._timer_stack.append(TimerData(
            name=name,
            start_time=start_time,
            parent_name=self._timer_stack[-1].log_name if self._timer_stack else None
        ))

    def end(self, name, end_time=None, bucket=None):
        """End a stopwatch span (must match latest started span)
        Arguments:
            name:
                Name of the scope that's completed. Must match the latest start()
            end_time:
                Time (s) at which the scope completed if set (if not use the current time)
            bucket:
                optional enum.Enum value describing bucket for additional reporting.
                For example, you might bucket all database queries together to see
                overall how much time is spent in databases.
        """

        if not end_time:
            end_time = self._time_func()

        tr_data = self._pop_stack(name)
        tr_data.end_time = end_time
        log_name = tr_data.log_name

        # Aggregate into a single bucket per concatenated log name. This makes sure that code like
        # the following code stopwatches as expected.
        #
        # with StopWatch.timer('cool_loop_time'):
        #     for x in cool_loop:
        #         cool_stuff(x)
        tr_delta_ms = max((tr_data.end_time - tr_data.start_time) * 1000.0, 0.001)
        if log_name in self._reported_values:
            self._reported_values[log_name][0] += tr_delta_ms
            self._reported_values[log_name][1] += 1
        else:
            self._reported_values[log_name] = [tr_delta_ms, 1, bucket]

        # go through slow tags and add them as tags if enough time has passed
        if not self._timer_stack:
            tr_data.trace_annotations.extend(self._root_annotations)

            threshold_s = tr_delta_ms / 1000.0
            for slowtag, timelimit in self._slow_annotations.items():
                if timelimit <= threshold_s:
                    tr_data.trace_annotations.append(
                        TraceAnnotation(slowtag, '1', tr_data.end_time)
                    )

        if self._should_trace_timer(log_name, tr_delta_ms):
            tr_data.parent_span_id = self._timer_stack[-1].span_id if self._timer_stack else None
            self._reported_traces.append(tr_data)

        # report stopwatch values once the final 'end' call has been made
        if not self._timer_stack:
            agg_report = AggregatedReport(self._reported_values, tr_data)
            # Stash information internally
            self._last_trace_report = self._reported_traces
            self._last_aggregated_report = agg_report
            # Hit callbacks
            self._export_tracing_func(reported_traces=self._reported_traces)
            self._export_aggregated_timers_func(aggregated_report=agg_report)

            self._reset()  # Clear out stats to prevent duplicate reporting

    def cancel(self, name):
        """Cancels a stopwatch span (must match latest started span).
        This span will not show up in any reports.
        Arguments:
            name:
                Name of the scope that's being cancelled. Must match the latest start().
        """
        self._pop_stack(name, end_type='cancel')
        self._cancelled_spans[name] = 1

    def add_annotation(self, key, value='1', event_time=None):
        """Add an annotation to the root scope. Note that we don't do this directly
        in order to support this case:

        if x > 1000:
            sw.add_annotation('big_work')
        with sw.timer('root'):
            do_work(x)

        """
        if event_time is None:
            event_time = self._time_func()
        self._root_annotations.append(
            TraceAnnotation(key, value, event_time)
        )

    def add_span_annotation(self, key, value='1', event_time=None):
        """Add an annotation to the current scope"""
        if event_time is None:
            event_time = self._time_func()
        self._timer_stack[-1].trace_annotations.append(
            TraceAnnotation(key, value, event_time)
        )

    def add_slow_annotation(self, tag, timelimit):
        """add annotation that will only be used if root scope takes longer than
        timelimit amount of seconds
        Arguments:
            tag: String tag name for the slowtag
            timelimit: Lower bound for the root scope after which tag is applied
        """
        self._slow_annotations[tag] = timelimit

    def get_last_trace_report(self):
        """Returns the last trace report from when the last root_scope completed"""
        return self._last_trace_report

    def get_last_aggregated_report(self):
        """Returns the last aggregated report and tags as a 2-tuple"""
        return self._last_aggregated_report

    def format_last_report(self):
        """Return formatted report from the last aggregated report. Simply calls
        format_last_report() on the get_last_aggregated_report()"""
        return format_report(self._last_aggregated_report)

    #################
    # Private methods
    #################

    def _pop_stack(self, name, end_type='end'):
        """Remove elements off the top of the timer stack until the element with name `name` is found.
        Return that element."""
        if not self._timer_stack:
            assert not self._strict_assert, \
                "StopWatch %s called but stack is empty: %s" % (end_type, name,)
            return

        tr_data = self._timer_stack.pop()
        assert (not self._strict_assert) or (tr_data.name == name), \
            "StopWatch %s: %s, does not match latest start: %s" % (end_type, name, tr_data.name)

        # if the top element on stack doesn't match "name", need to pop off things from the stack
        # till it matches to maximally negate the possible inconsistencies
        while name != tr_data.name and self._timer_stack:
            tr_data = self._timer_stack.pop()
        return tr_data

    def _should_trace_timer(self, log_name, delta_ms):
        """
        Helper method to determine if we should log the message or not.
        """
        if delta_ms < self.TRACING_MIN_NUM_MILLISECONDS:
            return False

        # Check if we have logged too many spans with the same full path already.
        # If yes, we should stop logging so we don't overload. E.g. if someone
        # is making for loop with 50k stopwatches, we will log only the first
        # MAX_REQUEST_TRACING_SPANS_FOR_PATH spans.

        return bool(log_name not in self._reported_values or
                    self._reported_values[log_name][1] <= self.MAX_REQUEST_TRACING_SPANS_FOR_PATH)



#!/usr/bin/env python2
# coding: utf-8

import unittest

from pykit import cluster
from pykit import ututil

dd = ututil.dd

_ec_config = {
    'ec': {
        'in_idc': [4, 2],
        'cross_idc': [2, 1],
        'ec_policy': 'lrc',
        'data_replica': 3

    }
}

_empty_bkg = {
    'bg_seq': 0,
    'config': {
        'ec': {
            'in_idc': [4, 2],
            'ec_policy': 'lrc',
            'cross_idc': [2, 1],
            'data_replica': 3
        }
    },
    'blocks': {
        '0000': {'is_del': 1, 'range': [None, None], 'type': 'd0', 'block_id': None, 'size': 0},
        '0001': {'is_del': 1, 'range': [None, None], 'type': 'd0', 'block_id': None, 'size': 0},
        '0002': {'is_del': 1, 'range': [None, None], 'type': 'd0', 'block_id': None, 'size': 0},
        '0003': {'is_del': 1, 'range': [None, None], 'type': 'd0', 'block_id': None, 'size': 0},

        '0004': {'is_del': 1, 'range': [None, None], 'type': 'dp', 'block_id': None, 'size': 0},
        '0005': {'is_del': 1, 'range': [None, None], 'type': 'dp', 'block_id': None, 'size': 0},

        '0100': {'is_del': 1, 'range': [None, None], 'type': 'd0', 'block_id': None, 'size': 0},
        '0101': {'is_del': 1, 'range': [None, None], 'type': 'd0', 'block_id': None, 'size': 0},
        '0102': {'is_del': 1, 'range': [None, None], 'type': 'd0', 'block_id': None, 'size': 0},
        '0103': {'is_del': 1, 'range': [None, None], 'type': 'd0', 'block_id': None, 'size': 0},

        '0104': {'is_del': 1, 'range': [None, None], 'type': 'dp', 'block_id': None, 'size': 0},
        '0105': {'is_del': 1, 'range': [None, None], 'type': 'dp', 'block_id': None, 'size': 0},

        '0200': {'is_del': 1, 'range': [None, None], 'type': 'x0', 'block_id': None, 'size': 0},
        '0201': {'is_del': 1, 'range': [None, None], 'type': 'x0', 'block_id': None, 'size': 0},
        '0202': {'is_del': 1, 'range': [None, None], 'type': 'x0', 'block_id': None, 'size': 0},
        '0203': {'is_del': 1, 'range': [None, None], 'type': 'x0', 'block_id': None, 'size': 0},

        '0204': {'is_del': 1, 'range': [None, None], 'type': 'xp', 'block_id': None, 'size': 0},
        '0205': {'is_del': 1, 'range': [None, None], 'type': 'xp', 'block_id': None, 'size': 0},

    },
    'idcs': ['a', 'b', 'c'],
    'block_group_id': 'g000640000000123'
}


class TestClusterBlockGroup(unittest.TestCase):
    def test_parse_and_print(self):
        block_group_id = 'g000640000000123'

        bgid = cluster.BlockGroupID(64, 123)
        self.assertEqual(block_group_id, str(bgid))

        bgid = cluster.BlockGroupID.parse(block_group_id)

        self.assertEqual(64, bgid.block_size)
        self.assertEqual(123, bgid.seq)

        self.assertEqual(block_group_id, str(bgid))
        self.assertEqual(block_group_id, '{0}'.format(bgid))
        self.assertEqual(
            "_BlockGroupID(block_size=64, seq=123)", repr(bgid))

        # test invalid input
        block_group_id_invalid = 'g00064000000012345'
        self.assertRaises(cluster.BlockGroupIDError,
                          cluster.BlockGroupID.parse, block_group_id_invalid)

    def test_tostr(self):
        block_group_id = 'g000640000000123'
        bgid = cluster.BlockGroupID.parse(block_group_id)
        self.assertEqual(block_group_id, str(bgid))
        self.assertEqual(block_group_id, bgid.tostr())

    def test_make_and_new(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)

        self.assertDictEqual(_empty_bkg, block_group)

        self.assertDictEqual(_empty_bkg, cluster.BlockGroup(block_group))

    def test_get_block(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)

        _, block = block_group.get_block(block_index='0000')

        self.assertDictEqual({
            'block_id': None,
            'is_del': 1,
            'range': [None, None],
            'type': 'd0',
            'size': 0
        }, block)

        _, block = block_group.get_block(block_index='9999')
        self.assertIsNone(block)

        with self.assertRaises(cluster.BlockNotFoundError):
            block_group.get_block(block_index='9999', raise_error=True)

        new_block = {
            'block_id': cluster.BlockID('d0',
                                        'g000640000000123',
                                        '0000',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd0',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        block_group.replace_block(new_block, block_index='0000')
        _, block = block_group.get_block(block_id=new_block['block_id'])
        self.assertDictEqual(new_block, block)

        _, block = block_group.get_block(block_id='d0g0006400000001230000c62d8736c72800020000000002')
        self.assertIsNone(block)

        with self.assertRaises(cluster.BlockNotFoundError):
            block_group.get_block(block_id='d0g0006400000001230000c62d8736c72800020000000002', raise_error=True)

        with self.assertRaises(cluster.BlockIDError):
            block_group.get_block(block_id='d0g0006400000001230000c62d2')

    def test_mark_delete_block(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)
        new_block = {
            'block_id': cluster.BlockID('d0',
                                        'g000640000000123',
                                        '0000',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd0',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        block_group.replace_block(new_block, block_index='0000')
        block_group.mark_delete_block(block_index='0000')
        _, block = block_group.get_block(block_index='0000')

        self.assertEqual(1, block['is_del'])

    def test_replace_block(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)
        new_block = {
            'block_id': cluster.BlockID('d0',
                                        'g000640000000123',
                                        '0000',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd0',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        block_group.replace_block(new_block, block_index='0000')
        _, block = block_group.get_block(block_index='0000')
        self.assertEqual(0, block['is_del'])

    def test_get_free_block_index(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)

        new_block = {
            'block_id': cluster.BlockID('d0',
                                        'g000640000000123',
                                        '0000',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd0',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        block_group.replace_block(new_block, block_index='0000')

        self.assertDictEqual({'a': ['0001', '0002', '0003'],
                              'b': ['0100', '0101', '0102', '0103']},
                             block_group.get_free_block_indexes('d0'))

        self.assertDictEqual({'a': ['0004', '0005'],
                              'b': ['0104', '0105']},
                             block_group.get_free_block_indexes('dp'))

        self.assertDictEqual({'c': ['0200', '0201', '0202', '0203'], },
                             block_group.get_free_block_indexes('x0'))

        self.assertDictEqual({'c': ['0204', '0205'], },
                             block_group.get_free_block_indexes('xp'))

    def test_get_block_type(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)

        self.assertEqual('d0', block_group.get_block_type('0000'))
        self.assertEqual('dp', block_group.get_block_type('0004'))
        self.assertEqual('d1', block_group.get_block_type('0006'))

        self.assertEqual('d0', block_group.get_block_type('0100'))
        self.assertEqual('dp', block_group.get_block_type('0104'))
        self.assertEqual('d1', block_group.get_block_type('0106'))

        self.assertEqual('x0', block_group.get_block_type('0200'))
        self.assertEqual('xp', block_group.get_block_type('0204'))

        with self.assertRaises(cluster.BlockTypeNotSupported):
            block_group.get_block_type('0209')

    def test_empty_block(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)

        new_block = {
            'block_id': cluster.BlockID('d0',
                                        'g000640000000123',
                                        '0000',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd0',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        block_group.replace_block(new_block, block_index='0000')
        block_group['blocks']['0000'] = block_group.empty_block('0000')
        _, block = block_group.get_block(block_index='0000')

        self.assertDictEqual({
            'block_id': None,
            'is_del': 1,
            'range': [None, None],
            'type': 'd0',
            'size': 0
        }, block)

    def test_get_block_idc(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)

        self.assertEqual('a', block_group.get_block_idc(block_index='0000'))
        self.assertEqual('b', block_group.get_block_idc(block_index='0100'))
        self.assertEqual('c', block_group.get_block_idc(block_index='0200'))

        d0 = {
            'block_id': cluster.BlockID('d0',
                                        'g000640000000123',
                                        '0000',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd0',
            'range': ['0a', '0b'],
            'is_del': 0
        }
        block_group.replace_block(d0, block_index='0000')
        self.assertEqual('a', block_group.get_block_idc(block_id=d0['block_id']))

    def test_get_replica_block_index(self):
        block_group = cluster.BlockGroup.make('g000640000000123', ['a', 'b', 'c'], _ec_config)
        self.assertEqual(['0006', '0010'], block_group.get_replica_block_index(block_index='0000'))
        self.assertEqual(['0000', '0010'], block_group.get_replica_block_index(block_index='0006'))
        self.assertEqual(['0000', '0006'], block_group.get_replica_block_index(block_index='0010'))

        with self.assertRaises(cluster.BlockTypeNotSupportReplica):
            block_group.get_replica_block_index(block_index='0004')

        with self.assertRaises(cluster.BlockTypeNotSupportReplica):
            block_group.get_replica_block_index(block_index='0200')

        with self.assertRaises(cluster.BlockTypeNotSupportReplica):
            block_group.get_replica_block_index(block_index='0204')

        d0 = {
            'block_id': cluster.BlockID('d0',
                                        'g000640000000123',
                                        '0000',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd0',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        d1 = {
            'block_id': cluster.BlockID('d1',
                                        'g000640000000123',
                                        '0006',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd1',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        d2 = {
            'block_id': cluster.BlockID('d2',
                                        'g000640000000123',
                                        '0010',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd2',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        dp = {
            'block_id': cluster.BlockID('dp',
                                        'g000640000000123',
                                        '0004',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'dp',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        x0 = {
            'block_id': cluster.BlockID('x0',
                                        'g000640000000123',
                                        '0200',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'x0',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        xp = {
            'block_id': cluster.BlockID('xp',
                                        'g000640000000123',
                                        '0204',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'xp',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        block_group.replace_block(d0, block_index='0000')

        block_group['blocks']['0006'] = d1
        block_group['blocks']['0010'] = d2

        block_group['blocks']['0004'] = dp
        block_group['blocks']['0200'] = x0
        block_group['blocks']['0204'] = xp

        self.assertEqual(['0006', '0010'], block_group.get_replica_block_index(block_id=d0['block_id']))
        self.assertEqual(['0000', '0010'], block_group.get_replica_block_index(block_id=d1['block_id']))
        self.assertEqual(['0000', '0006'], block_group.get_replica_block_index(block_id=d2['block_id']))

        with self.assertRaises(cluster.BlockTypeNotSupportReplica):
            block_group.get_replica_block_index(block_id=dp['block_id'])

        with self.assertRaises(cluster.BlockTypeNotSupportReplica):
            block_group.get_replica_block_index(block_id=x0['block_id'])

        with self.assertRaises(cluster.BlockTypeNotSupportReplica):
            block_group.get_replica_block_index(block_id=xp['block_id'])

    def test_block_group_init(self):
        block_group = cluster.BlockGroup(_empty_bkg)
        new_block = {
            'block_id': cluster.BlockID('d0',
                                        'g000640000000123',
                                        '0000',
                                        cluster.DriveID.parse('c62d8736c7280002'),
                                        1).tostr(),
            'size': 1000,
            'type': 'd0',
            'range': ['0a', '0b'],
            'is_del': 0
        }

        block_group.replace_block(new_block, block_index='0000')
        self.assertNotEqual(_empty_bkg, block_group)

        block_group = cluster.BlockGroup(**_empty_bkg)
        block_group.replace_block(new_block, block_index='0000')
        self.assertNotEqual(_empty_bkg, block_group)

        block_group = cluster.BlockGroup(_empty_bkg, **_empty_bkg)
        block_group.replace_block(new_block, block_index='0000')
        self.assertNotEqual(_empty_bkg, block_group)



import logging
import os
import time

from pykit import utfjson
from pykit.cgrouparch import model

logger = logging.getLogger(__name__)


def account_one_cgroup(slot_number, subsystem_model, cgroup_path,
                       cgroup_conf, result):
    account_value = subsystem_model['account'](cgroup_path)
    result['value'] = account_value

    sub_cgroup = cgroup_conf.get('sub_cgroup')
    if sub_cgroup is None:
        return

    result['sub_cgroup'] = {}

    for sub_cgroup_name, sub_cgroup_conf in sub_cgroup.iteritems():
        sub_cgroup_path = os.path.join(cgroup_path, sub_cgroup_name)

        result['sub_cgroup'][sub_cgroup_name] = {}

        account_one_cgroup(slot_number, subsystem_model, sub_cgroup_path,
                           sub_cgroup_conf,
                           result['sub_cgroup'][sub_cgroup_name])


def _run(context, slot_number):
    arch_conf = context['arch_conf']['value']

    result = {}

    for subsystem_name, subsystem_arch_conf in arch_conf.iteritems():
        subsystem_model = model.subsystem[subsystem_name]

        cgroup_path = os.path.join(context['cgroup_dir'], subsystem_name)
        cgroup_conf = subsystem_arch_conf

        result[subsystem_name] = {}

        account_one_cgroup(slot_number, subsystem_model, cgroup_path,
                           cgroup_conf, result[subsystem_name])

    redis_client = context['redis_client']

    key_name = '%s/account/%d' % (context['redis_prefix'], slot_number)

    redis_client.set(key_name, utfjson.dump(result))
    redis_client.expire(key_name, context['redis_expire_time'])


def run(context):
    while True:
        try:
            start_ts = time.time()
            slot_number = int(round(start_ts))

            _run(context, slot_number)

            end_ts = time.time()

            logger.info('account at: %f, time used: %f' %
                        (start_ts, end_ts - start_ts))

            to_sleep = slot_number + 1 - end_ts
            if to_sleep <= 0:
                logger.error('account use too much time')
                to_sleep = 0.1

            time.sleep(to_sleep)

        except Exception as e:
            logger.exception('failed to account: %s' % repr(e))
            time.sleep(1)


def show(context, args):
    end_slot = args.get('end_slot', int(time.time()))
    nr_slot = args.get('nr_slot', 10)

    start_slot = args.get('start_slot', end_slot - nr_slot + 1)

    result = {}

    redis_client = context['redis_client']

    for slot_number in range(start_slot, end_slot + 1):
        key_name = '%s/account/%d' % (context['redis_prefix'], slot_number)

        value_str = redis_client.get(key_name)
        if value_str is None:
            continue

        result[slot_number] = utfjson.load(value_str)

    return result



import docker

from pykit import ututil

dd = ututil.dd

# default network config for unittest
net_config = {
    'network1': {
        'subnet': '192.168.52.0/24',
        'gateway': '192.168.52.254',
    }
}


def get_client():
    dcli = docker.Client(base_url='unix://var/run/docker.sock')
    return dcli


def does_container_exist(name):

    dcli = get_client()

    try:
        dcli.inspect_container(name)
        return True
    except docker.errors.NotFound:
        return False


def stop_container(*names):

    dcli = get_client()
    for name in names:
        try:
            dcli.stop(container=name)
        except Exception as e:
            dd(repr(e), ' while trying to stop docker container: ' + repr(name))


def remove_container(*names):

    dcli = get_client()

    for name in names:
        try:
            dcli.kill(name)
        except Exception as e:
            dd(repr(e) + ' while killing container: ' + repr(name))

        try:
            dcli.remove_container(name)
        except Exception as e:
            dd(repr(e) + ' while removing container: ' + repr(name))


def create_network():

    net_name = 'network1'
    dcli = get_client()

    try:
        dcli.inspect_network(net_name)
        return
    except docker.errors.NotFound as e:
        dd(repr(e))

    ipam_pool = docker.utils.create_ipam_pool(**net_config[net_name])

    ipam_config = docker.utils.create_ipam_config(
        pool_configs=[ipam_pool]
    )

    dcli.create_network(net_name, driver="bridge", ipam=ipam_config)


def start_container(name, image,
                    ip=None,
                    command='',
                    port_bindings=None,
                    volume_bindings=None,
                    env=None):

    net_name = 'network1'
    dcli = get_client()

    kwargs = {}
    _config_kwargs = {}
    if env is not None:
        kwargs['environment'] = env

    if port_bindings is not None:
        kwargs['ports'] = port_bindings.keys()
        _config_kwargs['port_bindings'] = port_bindings

    if volume_bindings is not None:
        volumes = []
        for bind in volume_bindings:
            volumes.append(bind.split(':')[1])

        kwargs['volumes'] = volumes
        _config_kwargs['binds'] = volume_bindings

    kwargs['host_config'] = dcli.create_host_config(
        **_config_kwargs)

    if ip is not None:
        net_cfg = dcli.create_networking_config({
            net_name: dcli.create_endpoint_config(ipv4_address=ip,)
        })
        kwargs['networking_config'] = net_cfg

    if not does_container_exist(name):
        dcli.create_container(
            name=name,
            image=image,
            command=command,
            **kwargs
        )

    dcli.start(container=name)


def pull_image(image):

    dcli = get_client()
    rst = dcli.images(image)
    if len(rst) > 0:
        dd(image + ' is ready')
        dd(rst)
        return

    # 'daocloud.io/zookeeper:3.4.10' --> ('daocloud.io/zookeeper', '3.4.10')
    rst = dcli.pull(*image.split(':'))
    dd(rst)



#!/usr/bin/env python2
# coding: utf-8

import unittest

from pykit import cluster
from pykit import ututil

dd = ututil.dd


class TestClusterRegion(unittest.TestCase):

    def test_init(self):

        region_cases = [
            [{},
                {'idc': None, 'range': [None, None], 'levels': []}],
            [{'range': ['a', 'b'], 'idc': '.bei'},
                {'idc': '.bei', 'range': ['a', 'b'], 'levels': []}],
            [{'levels': [[['a', 'b', 1]], [['c', 'd', 2]]]},
                {'idc': None, 'range': [None, None], 'levels': [[['a', 'b', 1]], [['c', 'd', 2]]]}],
            [{'range': ['a', 'z'], 'levels': [[['a', 'b', 1], ['b', 'c', 3]]]},
                {'idc': None, 'range': ['a', 'z'], 'levels': [[['a', 'b', 1], ['b', 'c', 3]]]}],
        ]

        for case, excepted in region_cases:

            region = cluster.Region(case)

            case['idc'] = 'beijing'
            case['test'] = {'a': 1}
            if 'levels' in case:
                case['levels'].append(['d', 'e', 5])

            self.assertEqual(excepted, region)

        region_cases_argkv = [
            ([[['a', 'b', 1], ['c', 'd', 2]]],
                {'idc': None, 'range': [None, None], 'levels': [[['a', 'b', 1], ['c', 'd', 2]]]}),
            (['a', 'z'],
                {'idc': None, 'range': ['a', 'z'], 'levels': []}),
            ([],
                {'idc': None, 'range': ['a', 'z'], 'levels': [[['a', 'b', 1], ['c', 'd', 2]], [['f', 'g', 5]]]}),
        ]

        region = cluster.Region(levels=region_cases_argkv[0][0])
        region_cases_argkv[0][0].append([['f', 'g', 5]])
        self.assertEqual(region_cases_argkv[0][1], region)

        region = cluster.Region(range=region_cases_argkv[1][0])
        self.assertEqual(region_cases_argkv[1][1], region)

        region = cluster.Region(
            levels=region_cases_argkv[0][0], range=region_cases_argkv[1][0])
        self.assertEqual(region_cases_argkv[2][1], region)

    def test_move_down(self):

        region_levels = [
            [['aa', 'ee', 1], ['hh', 'hz', 2], ['pp', 'zz', 3], ['zz', None, 4]],
            [['cf', 'cz', 5], ['mm', 'oo', 6], ['oo', 'qq', 7]],
            [['aa', 'bb', 8], ['cc', 'cd', 9], ['ee', 'ff', 10]],
            [['aa', 'ab', 11], ['az', 'bb', 12], ['za', None, 13]],
            [['d', 'fz', 14]],
        ]

        excepted_region_levels = [
            [['aa', 'ee', 1], ['ee', 'ff', 10], ['hh', 'hz', 2], [
                'mm', 'oo', 6], ['pp', 'zz', 3], ['zz', None, 4]],
            [['aa', 'bb', 8], ['cc', 'cd', 9], ['cf', 'cz', 5], [
                'd', 'fz', 14], ['oo', 'qq', 7], ['za', None, 13]],
            [['aa', 'ab', 11], ['az', 'bb', 12]],
        ]

        excepted_moved_blocks = [
            (1, 0, ['mm', 'oo', 6]),
            (2, 1, ['aa', 'bb', 8]),
            (2, 1, ['cc', 'cd', 9]),
            (2, 0, ['ee', 'ff', 10]),
            (3, 2, ['aa', 'ab', 11]),
            (3, 2, ['az', 'bb', 12]),
            (3, 1, ['za', None, 13]),
            (4, 1, ['d', 'fz', 14]),
        ]

        region = cluster.Region(levels=region_levels)
        moved_blocks = region.move_down()

        self.assertEqual(excepted_moved_blocks, moved_blocks)
        self.assertEqual(excepted_region_levels, region['levels'])

        region_levels = [
            [['aa', 'ee', 1], ['ee', 'ff', 10], ['hh', 'hz', 2]],
            [['aa', 'yy', 8]]
        ]

        region = cluster.Region(levels=region_levels)
        moved_blocks = region.move_down()

        self.assertEqual([], moved_blocks)
        self.assertEqual(region_levels, region['levels'])

    def test_find_merge(self):

        region_levels_cases = [
            [
                [
                    [['aa', 'ee', {'size': 8}], ['ee', 'ff', {'size': 16}], [
                        'pp', 'zz', {'size': 8}], ['zz', None, {'size': 4}]],
                    [['aa', 'pz', {'size': 4}], ['qq', 'zz', {'size': 8}]]
                ],
                (1, ['qq', 'zz', {'size': 8}], [['pp', 'zz', {'size': 8}]])
            ],
            [
                [
                    [['aa', 'ee', {'size': 8}], ['ee', 'ff', {
                        'size': 8}], ['hh', 'hz', {'size': 8}]],
                    [['mm', 'yy', {'size': 8}]]
                ],
                None
            ]
        ]

        for levels, excepted in region_levels_cases:
            region = cluster.Region(levels=levels)
            res = region.find_merge()

            self.assertEqual(excepted, res)

    def test_list_block_ids(self):
        region_levels = [
            [['aa', 'ee', {'block_id': 1}], ['hh', 'zz', {'block_id': 2}]],
            [['ea', 'ff', {'block_id': 4}], ['mm', 'yy', {'block_id': 5}]],
        ]

        cases = (
                (None, [1, 2, 4, 5]),
                (3,    [4, 5]),
                (5,     [5]),
                (6,     []),
        )

        region = cluster.Region(levels=region_levels)

        for bid, excepted in cases:
            block_ids = region.list_block_ids(start_block_id=bid)
            self.assertEqual(excepted, block_ids)

    def test_replace_block_id(self):
        region_levels = [
            [['aa', 'ee', {'block_id': 1}], ['hh', 'zz', {'block_id': 2}]],
            [['ea', 'ff', {'block_id': 4}], [
                'mm', 'yy', {'block_id': 5}]]
        ]

        excepted_region_levels = [
            [['aa', 'ee', {'block_id': 1}], ['hh', 'zz', {'block_id': 2}]],
            [['ea', 'ff', {'block_id': 3}], [
                'mm', 'yy', {'block_id': 5}]]
        ]

        region = cluster.Region(levels=region_levels)

        region.replace_block_id(4, 3)
        self.assertEqual(excepted_region_levels, region['levels'])

        self.assertRaises(cluster.BlockNotInRegion,
                          region.replace_block_id, 7, 9)



import random
import unittest

from pykit import priorityqueue
from pykit import threadutil
from pykit import ututil

dd = ututil.dd


class TestConsumedQueue(unittest.TestCase):

    def test_init(self):

        cq = priorityqueue.Producer(1, 1, [])
        self.assertEqual(1, cq.priority)
        self.assertEqual(0, cq.consumed)
        self.assertEqual(priorityqueue.default_priority / cq.priority,
                         cq.item_cost)

    def test_priority(self):
        self.assertRaises(ValueError, priorityqueue.Producer, 1, 0, [])
        self.assertRaises(ValueError, priorityqueue.Producer, 1, -1, [])
        self.assertRaises(ValueError, priorityqueue.Producer, 1, '', [])

    def test_consumed(self):
        a = priorityqueue.Producer(1, 1, [1, 2])
        b = priorityqueue.Producer(1, 10, range(20))

        self.assertEqual(1, a.get())
        self.assertEqual(2, a.get())
        self.assertRaises(priorityqueue.Empty, a.get)

        for i in range(20):
            b.get()

        self.assertEqual(a.consumed, b.consumed)

    def test_lt_init(self):
        a = priorityqueue.Producer(1, 1, [1, 2])
        b = priorityqueue.Producer(1, 1, [1, 2])

        self.assertFalse(a < b, 'both empty')
        self.assertFalse(b < a, 'both empty')

        a.get()
        b.get()

        self.assertFalse(a < b, 'both empty')
        self.assertFalse(b < a, 'both empty')

    def test_lt(self):
        a = priorityqueue.Producer(1, 1, range(100))
        b = priorityqueue.Producer(1, 10, range(100))

        # 0 0, but b has smaller item_cost
        self.assertFalse(a < b, 'b has higher priority')
        self.assertTrue(b < a, 'b has higher priority')

        # 0 0.1
        b.get()
        self.assertTrue(a < b,  'b consumed 1')
        self.assertFalse(b < a,  'b consumed 1')

        # 1 0.1
        a.get()
        self.assertFalse(a < b, 'a consumed 1')
        self.assertTrue(b < a,  'a consumed 1')

        # 1 0.2
        b.get()
        self.assertFalse(a < b,  'a b consumed 1')
        self.assertTrue(b < a,  'a b consumed 1')

        # 1 1
        for _ in range(8):
            b.get()
        self.assertFalse(a < b,  'a b consumed 1')
        self.assertTrue(b < a,  'a consume 1 b consumed 10')

        # 1 1.1
        b.get()
        self.assertTrue(a < b,  'a b consumed 1')
        self.assertFalse(b < a,  'a b consumed 1')


class TestPriorityQueue(unittest.TestCase):

    def test_get(self):
        producers = (
            # id, priority, iterable
            (1, 1, [1] * 10),
            (2, 2, [2] * 10),
            (3, 3, [3] * 10),
        )
        expected = [3,  # .3    0  0.0
                    2,  # .3   .5  0.0
                    1,  # .3   .5  1.0
                    3,  # .6   .5  1.0
                    2,  # .6  1.0  1.0
                    3,  # 1.0  1.0  1.0
                    3,  # 1.3  1.0  1.0
                    2,  # 1.3  1.5  1.0
                    1,  # 1.3  1.5  2.0
                    3,  # 1.6  1.5  2.0
                    2,  # 1.6  2.0  2.0
                    3,  # 2.0  2.0  2.0
                    ]
        pq = priorityqueue.PriorityQueue()
        for pid, prio, itr in producers:
            pq.add_producer(pid, prio, itr)

        dd(pq)
        rst = []
        for _ in expected:
            val = pq.get()
            rst.append(val)
        self.assertEqual(expected, rst)

    def test_add_queue(self):
        pq = priorityqueue.PriorityQueue()
        pq.add_producer(1, 1, [1, 1])
        self.assertEqual(1, pq.get())

        # re-add should update iterable
        pq.add_producer(1, 1, [2, 2])
        self.assertEqual(2, pq.get())

        # new queue with higher priority comes first
        pq.add_producer(2, 10, [10, 10])
        self.assertEqual(10, pq.get())

    def test_add_queue_update_priority(self):
        pq = priorityqueue.PriorityQueue()

        pq.add_producer(1, 1, [1, 1, 1])
        self.assertEqual(1, pq.get())

        pq.add_producer(2, 2, [2, 2, 2])
        self.assertEqual(2, pq.get())
        self.assertEqual(2, pq.get())

        # re-add should update priority, iterable and adjust heap
        pq.add_producer(1, 10, [10, 10, 10])
        dd(pq)
        dd(pq.consumable_heap)

        self.assertEqual(10, pq.get())

    def test_remove_queue(self):
        pq = priorityqueue.PriorityQueue()
        pq.add_producer(1, 1, [1, 1])
        pq.remove_producer(1)

        self.assertRaises(priorityqueue.Empty, pq.get)
        pq.add_producer(1, 1, [1, 1])
        pq.add_producer(10, 10, [10, 10])
        self.assertEqual(10, pq.get())

        pq.remove_producer(10)
        self.assertEqual(1, pq.get())

    def test_remove_queue_inexistent(self):
        pq = priorityqueue.PriorityQueue()
        self.assertRaises(KeyError, pq.remove_producer, 1)
        pq.remove_producer(1, ignore_not_found=True)

    def test_bench(self):

        pq = priorityqueue.PriorityQueue()

        ntimes = 10240
        nq = 1024
        n_thread = 3
        ths = []

        def worker():
            for _ in range(ntimes):
                pq.get()

        for i in range(1, nq + 1):
            pq.add_producer(i, i, yield_forever())

        with ututil.Timer() as t:
            for i in range(n_thread):
                th = threadutil.start_daemon_thread(worker)
                ths.append(th)

            for th in ths:
                th.join()

            us_per_call = t.spent() / ntimes / n_thread * 1000 * 1000
            dd(us_per_call, 'us/call')

        self.assertLess(us_per_call, 50)

    def test_concurrent(self):
        pq = priorityqueue.PriorityQueue()
        ntimes = 10240
        nq = 3
        n_thread = 3
        ths = []

        def worker():
            for _ in range(ntimes):
                pq.get()

        for i in range(1, nq + 1):
            pq.add_producer(i, i, yield_forever())

        for i in range(n_thread):
            th = threadutil.start_daemon_thread(worker)
            ths.append(th)

        for th in ths:
            th.join()

        consumed = []
        got = 0
        for i in range(1, nq + 1):
            q = pq.producer_by_id[i]
            consumed.append(q.consumed)
            dd('get:', q.stat)
            got += q.stat['get']

        self.assertEqual(ntimes * n_thread, got)

        dd('consumed:', consumed)
        self.assertAlmostEqual(consumed[0], consumed[1])
        self.assertAlmostEqual(consumed[1], consumed[2])


def yield_forever():
    i = random.randint(1, 1000000)
    while True:
        i += 1
        yield (i * 2654435769) & 0xffffffff



#!/usr/bin/env python2
# coding: utf-8

import errno
import re
import types
import yaml


cached_schemas = {}

BOOL = 'bool'
INTEGER = 'integer'
FLOAT = 'float'
STRING_NUMBER = 'string_number'
STRING = 'string'
ARRAY = 'array'
DICT = 'dict'
ANY = 'any'
# SCHEMA type should only be used by schema validator.
SCHEMA = 'schema'


TYPE_MAPPING = {
    BOOL: [types.BooleanType],
    INTEGER: [types.IntType, types.LongType],
    FLOAT: [types.IntType, types.LongType, types.FloatType],
    STRING_NUMBER: list(types.StringTypes),
    STRING: list(types.StringTypes),
    ARRAY: [types.ListType],
    DICT: [types.DictType],
    ANY: [types.BooleanType, types.IntType, types.LongType, types.FloatType,
          types.DictType, types.ListType] + list(types.StringTypes),

    SCHEMA: [types.DictType]
}

schema_keywords = {
    'checker': {
        'element_type': 'dict',
        'required': True,
        'scope': ['root', 'sub_schema'],
        'type': ['dict', 'array']
    },

    'element_checker': {
        'element_type': 'dict',
        'scope': ['checker', 'element_checker', 'value_checker'],
        'type': ['dict', 'array']
    },

    'enum': {
        'allowed_scope_types': ['string', 'integer', 'float', 'string_number'],
        'element_type': 'inherited',
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'array'
    },

    'fixed_length': {
        'allowed_scope_types': ['string', 'array', 'dict'],
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'integer'
    },

    'key_checker': {
        'element_type': 'dict',
        'scope': ['checker', 'element_checker', 'value_checker'],
        'type': ['dict', 'array']
    },

    'max': {
        'allowed_scope_types': ['integer', 'float', 'string_number'],
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'float'
    },

    'max_length': {
        'allowed_scope_types': ['string', 'array', 'dict'],
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'integer'
    },

    'min': {
        'allowed_scope_types': ['integer', 'float', 'string_number'],
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'float'
    },

    'min_length': {
        'allowed_scope_types': ['string', 'array', 'dict'],
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'integer'
    },

    'not': {
        'allowed_scope_types': ['string', 'integer', 'float', 'string_number'],
        'element_type': 'inherited',
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'array'
    },

    'regexp': {
        'allowed_scope_types': ['string'],
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'string'
    },

    'required': {
        'scope': ['root', 'sub_schema'],
        'type': 'bool'
    },

    'sub_schema': {
        'allowed_scope_types': ['dict'],
        'scope': ['checker', 'element_checker'],
        'type': 'schema'
    },

    'type': {
        'required': True,
        'scope': ['checker', 'element_checker', 'key_checker', 'value_checker'],
        'type': 'string'
    },

    'value_checker': {
        'element_type': 'dict',
        'scope': ['checker', 'element_checker', 'value_checker'],
        'type': ['dict', 'array']
    },
}


class SchemaCheckerError(Exception):
    pass


class SchemaNotFoundError(SchemaCheckerError):
    pass


class InvalidSchemaError(SchemaCheckerError):
    pass


class InvalidArgumentError(SchemaCheckerError):
    pass


class UnexpectedArgumentError(InvalidArgumentError):
    pass


class LackArgumentError(InvalidArgumentError):
    pass


class InvalidTypeError(InvalidArgumentError):
    pass


class OutOfRangeError(InvalidArgumentError):
    pass


class NotInEnumError(InvalidArgumentError):
    pass


class InvalidValueError(InvalidArgumentError):
    pass


class InvalidLengthError(InvalidArgumentError):
    pass


class PatternNotMatchError(InvalidArgumentError):
    pass


class ArgumentChecker(object):
    """
    Validate the argument based on the type definition in the schema.
    """

    any_checker = {'type': 'any'}

    def __init__(self, argument_name, value, checker):
        self.arg_name = argument_name
        self.val = value

        if type(checker) is list:
            checkers = checker
        else:
            checkers = [checker]

        self.checker = self.pick_checker_by_type(checkers)

    def pick_checker_by_type(self, checkers):
        checker_by_type = {c['type']: c for c in checkers}

        # Prefer string_number type if the val is a string number.
        if type(self.val) in TYPE_MAPPING['string']:
            if (_is_string_number(self.val) and
                    'string_number' in checker_by_type):
                return checker_by_type['string_number']

            elif 'string' in checker_by_type:
                return checker_by_type['string']

        for c in checkers:
            if type(self.val) in TYPE_MAPPING[c['type']]:
                return c

        types = [c['type'] for c in checkers]
        raise InvalidTypeError(self.arg_name, self.val, types)

    def check(self):
        func_name = 'check_{t}'.format(t=self.checker['type'])
        check_func = getattr(self, func_name)
        check_func()

    def _check_fixed_length(self):
        fixed_length = self.checker.get('fixed_length', None)
        if fixed_length is None:
            return

        if len(self.val) != fixed_length:
            raise InvalidLengthError(self.arg_name, self.val, fixed_length)

    def _check_length_range(self):
        min_len = self.checker.get('min_length', None)
        max_len = self.checker.get('max_length', None)
        val_len = len(self.val)

        if min_len is not None and val_len < min_len:
            raise InvalidLengthError(self.arg_name, self.val, min_len)

        if max_len is not None and val_len > max_len:
            raise InvalidLengthError(self.arg_name, self.val, max_len)

    def _check_length(self):
        self._check_fixed_length()
        self._check_length_range()

    def _check_invalid_values(self):
        invalid_vals = self.checker.get('not', None)
        if invalid_vals is None:
            return

        if self.val in invalid_vals:
            raise InvalidValueError(self.arg_name, self.val, invalid_vals)

    def _check_enums(self):
        enums = self.checker.get('enum', None)
        if enums is None:
            return

        if self.val not in enums:
            raise NotInEnumError(self.arg_name, self.val, enums)

    def _check_number_range(self):
        val = float(self.val)
        min_number = self.checker.get('min', None)
        max_number = self.checker.get('max', None)

        if min_number is not None and val < min_number:
            raise OutOfRangeError(self.arg_name, self.val, min_number)

        if max_number is not None and val > max_number:
            raise OutOfRangeError(self.arg_name, self.val, max_number)

    def _check_elements(self):
        elt_checker = self.checker.get('element_checker', None)
        if elt_checker is None:
            return

        for elt in self.val:
            elt_name = "{arg}'s element".format(arg=self.arg_name)
            ArgumentChecker(elt_name, elt, elt_checker).check()

    def _check_k_v(self):
        k_checker = self.checker.get('key_checker', None)
        v_checker = self.checker.get('value_checker', None)
        if k_checker is None and v_checker is None:
            return

        for k, v in self.val.iteritems():
            if k_checker is not None:
                k_name = "{arg}'s key".format(arg=self.arg_name)
                ArgumentChecker(k_name, k, k_checker).check()

            if v_checker is not None:
                v_name = "{arg}'s value".format(arg=self.arg_name)
                ArgumentChecker(v_name, v, v_checker).check()

    def _check_sub_schema(self):
        sub_schema = self.checker.get('sub_schema', None)
        if sub_schema is None:
            return

        check_arguments(self.val, sub_schema)

    def _check_regexp(self):
        regexp = self.checker.get('regexp')
        if regexp is None:
            return

        pattern = re.compile(regexp)
        if pattern.match(self.val) is None:
            raise PatternNotMatchError(self.arg_name, self.val, regexp)

    def check_any(self):
        pass

    def check_bool(self):
        pass

    def check_float(self):
        self._check_number_range()
        self._check_enums()
        self._check_invalid_values()

    def check_integer(self):
        self.check_float()

    def check_string_number(self):
        if not _is_string_number(self.val):
            raise InvalidTypeError(self.arg_name, self.val, ['string_number'])

        self.check_float()

    def check_string(self):
        self._check_enums()
        self._check_invalid_values()
        self._check_length()
        self._check_regexp()

    def check_array(self):
        self._check_length()
        self._check_elements()

    def check_dict(self):
        self._check_length()
        self._check_k_v()
        self._check_sub_schema()


def _is_string_number(val):
    try:
        float(val)
        return True
    except ValueError:
        return False


def _get_expected_types(keyword_type, scope_type):
    """
    Get the actual type of the keyword based on the defined keyword types, and
    the scope type that is only useful when the defined type is 'inherited'.
    """

    if keyword_type == 'inherited':
        assert scope_type is not None, ("Scope type should not be None "
                                        "when checking inherited value")
        return [scope_type]

    if type(keyword_type) is list:
        return keyword_type
    else:
        return [keyword_type]


def _validate_schema_in_scope(scope, schema):
    # Make sure required keywords exist.
    for keyword, definition in schema_keywords.iteritems():
        if (scope in definition['scope'] and definition.get('required') and
                keyword not in schema):
            raise InvalidSchemaError("'{k}' is required for scope '{s}'".
                                     format(k=keyword, s=scope))

    scope_type = schema.get('type')
    for keyword, val in schema.iteritems():
        if keyword not in schema_keywords:
            raise InvalidSchemaError("'{k}' is not a valid keyword".
                                     format(k=keyword))

        k_def = schema_keywords[keyword]
        if scope not in k_def['scope']:
            raise InvalidSchemaError("'{k}' is not in the correct scope {s}".
                                     format(k=keyword, s=k_def['scope']))

        allowed_scope_types = k_def.get('allowed_scope_types')
        if allowed_scope_types is not None and scope_type not in allowed_scope_types:
            raise InvalidSchemaError(
                "'{k}' can only be used when the scope type is one of {t}, "
                "got {real_t}".format(k=keyword, t=allowed_scope_types,
                                      real_t=scope_type))

        expected_types = _get_expected_types(k_def['type'], scope_type)
        _validate_type(scope_type, expected_types, keyword, val)


def _validate_type(scope_type, expected_types, keyword, val):
    """
    Validate the type of a keyword, and its sub keywords recursively.
    """
    keyword_type = None
    for t in expected_types:
        if type(val) in TYPE_MAPPING[t]:
            keyword_type = t
            break
    else:
        raise InvalidSchemaError("Expected types '{t}' for '{k}', got '{v}'".
                                 format(t=expected_types, k=keyword, v=val))

    if keyword_type == SCHEMA:
        # Validate sub schema.
        validate_schema({}, val)

    elif keyword_type == DICT:
        _validate_schema_in_scope(keyword, val)

    elif keyword_type == ARRAY:
        k_def = schema_keywords[keyword]
        expected_types = _get_expected_types(k_def['element_type'], scope_type)

        for element in val:
            _validate_type(scope_type, expected_types, keyword, element)


def validate_schema(args, schema):
    if schema is None or len(schema) == 0:
        raise InvalidSchemaError('Empty Schema')

    if not isinstance(args, dict):
        _validate_schema_in_scope('root', schema)
        return

    for _, schema in schema.iteritems():
        _validate_schema_in_scope('root', schema)


def read_schema(path):
    try:
        f = open(path)
    except IOError as e:
        if e.errno == errno.ENOENT:
            raise SchemaNotFoundError(path)
        else:
            raise

    content = f.read()

    try:
        return yaml.load(content)
    except yaml.error.YAMLError as e:
        raise InvalidSchemaError('{err} schema cannot be parsed'.format(
                                 err=repr(e)))


def check_arguments(args, schema):
    validate_schema(args, schema)

    if not isinstance(args, dict):
        ArgumentChecker('', args, schema['checker']).check()
        return

    for arg_name, sub_schema in schema.iteritems():
        if sub_schema.get('required', False) and arg_name not in args:
            raise LackArgumentError(arg_name)

    for arg_name, val in args.iteritems():
        if arg_name not in schema:
            raise UnexpectedArgumentError(arg_name, schema.keys())

        checker_schema = schema[arg_name]['checker']
        ArgumentChecker(arg_name, val, checker_schema).check()



import time

import daemonize

fn = '/tmp/bar'
pidfn = '/tmp/test_daemonize.pid'


def write_file(fn, cont):
    with open(fn, 'w') as f:
        f.write(cont)


def run():

    write_file(fn, 'bar-before')

    time.sleep(1)

    write_file(fn, 'bar-after')


daemonize.daemonize_cli(run, pidfn)



#!/usr/bin/env python2
# coding: utf-8

import copy
import logging
import threading
import time

logger = logging.getLogger(__name__)
_lock_mutex_update = threading.RLock()


class LRU(object):

    def __init__(self, capacity, timeout=60):

        self.lock = threading.RLock()
        self.capacity = capacity
        self.cleanup_threshold = int(capacity * 1.5)
        self.timeout = timeout
        self.size = 0
        self.items = {}
        self.head = {'next': None, 'pre': None}
        self.tail = self.head

    def __getitem__(self, key):

        with self.lock:
            now = int(time.time())
            item = self.items[key]

            if now > item['tm'] + self.timeout:
                self._del_item(item)
                raise KeyError('{k} is timeout'.format(k=key))

            self._move_to_tail(item)

            return item['val']

    def __setitem__(self, key, val):

        with self.lock:
            if key in self.items:
                item = self.items[key]
                item['val'] = val
                item['tm'] = int(time.time())

                self._move_to_tail(item)

            else:
                self.items[key] = {'key': key,
                                   'val': val,
                                   'pre': None,
                                   'next': None,
                                   'tm': int(time.time())}

                self._move_to_tail(self.items[key])

                self.size += 1

                if self.size > self.cleanup_threshold:
                    self._cleanup()

    def _remove_item(self, item):

        item['pre']['next'] = item['next']
        if item['next'] is not None:
            item['next']['pre'] = item['pre']
        else:
            self.tail = item['pre']

    def _move_to_tail(self, item):

        if item['pre'] is not None:
            self._remove_item(item)

        self.tail['next'] = item
        item['pre'] = self.tail
        item['next'] = None
        self.tail = item

    def _del_item(self, item):

        del self.items[item['key']]
        self._remove_item(item)
        self.size -= 1

    def _cleanup(self):

        while self.size > self.capacity:
            item = self.head['next']
            self._del_item(item)


class Cacheable(object):

    def __init__(self, capacity=1024 * 4, timeout=60,
                    is_deepcopy=True, mutex_update=False):

        self.lru = LRU(capacity, timeout)
        self.is_deepcopy = is_deepcopy
        self.mutex_update = mutex_update

    def _arg_str(self, args, argkv):

        argkv = [(k, v) for k, v in argkv.items()]
        argkv.sort()

        return str([args, argkv])

    def _cache_wrapper(self, fun):

        def func_wrapper(*args, **argkv):

            val = None
            generate_key = self._arg_str(args, argkv)

            try:
                val = self.lru[generate_key]
            except KeyError:
                if self.mutex_update:
                    with _lock_mutex_update:
                        try:
                            val = self.lru[generate_key]
                        except KeyError:
                            val = fun(*args, **argkv)
                            self.lru[generate_key] = val
                else:
                    val = fun(*args, **argkv)
                    self.lru[generate_key] = val

            if self.is_deepcopy:
                return copy.deepcopy(val)
            else:
                return val

        return func_wrapper


caches = {}


def cache(name, capacity=1024 * 4, timeout=60, is_deepcopy=True, mutex_update=False):

    c = caches.get(name)
    if c is None:
        c = Cacheable(capacity=capacity, timeout=timeout,
                      is_deepcopy=is_deepcopy, mutex_update=mutex_update)
        caches[name] = c

    return c._cache_wrapper



import sys
from pykit.p3json.test import PyTest


class TestScanstring(object):
    def test_scanstring(self):
        scanstring = self.json.decoder.scanstring
        if sys.maxunicode == 65535:
            self.assertEqual(
                scanstring(u'"z\U0001d120x"', 1, None, True),
                (u'z\U0001d120x', 6))
        else:
            self.assertEqual(
                scanstring(u'"z\U0001d120x"', 1, None, True),
                (u'z\U0001d120x', 5))

        self.assertEqual(
            scanstring('"\\u007b"', 1, None, True),
            (u'{', 8))

        self.assertEqual(
            scanstring('"A JSON payload should be an object or array, not a string."', 1, None, True),
            (u'A JSON payload should be an object or array, not a string.', 60))

        self.assertEqual(
            scanstring('["Unclosed array"', 2, None, True),
            (u'Unclosed array', 17))

        self.assertEqual(
            scanstring('["extra comma",]', 2, None, True),
            (u'extra comma', 14))

        self.assertEqual(
            scanstring('["double extra comma",,]', 2, None, True),
            (u'double extra comma', 21))

        self.assertEqual(
            scanstring('["Comma after the close"],', 2, None, True),
            (u'Comma after the close', 24))

        self.assertEqual(
            scanstring('["Extra close"]]', 2, None, True),
            (u'Extra close', 14))

        self.assertEqual(
            scanstring('{"Extra comma": true,}', 2, None, True),
            (u'Extra comma', 14))

        self.assertEqual(
            scanstring('{"Extra value after close": true} "misplaced quoted value"', 2, None, True),
            (u'Extra value after close', 26))

        self.assertEqual(
            scanstring('{"Illegal expression": 1 + 2}', 2, None, True),
            (u'Illegal expression', 21))

        self.assertEqual(
            scanstring('{"Illegal invocation": alert()}', 2, None, True),
            (u'Illegal invocation', 21))

        self.assertEqual(
            scanstring('{"Numbers cannot have leading zeroes": 013}', 2, None, True),
            (u'Numbers cannot have leading zeroes', 37))

        self.assertEqual(
            scanstring('{"Numbers cannot be hex": 0x14}', 2, None, True),
            (u'Numbers cannot be hex', 24))

        self.assertEqual(
            scanstring('[[[[[[[[[[[[[[[[[[[["Too deep"]]]]]]]]]]]]]]]]]]]]', 21, None, True),
            (u'Too deep', 30))

        self.assertEqual(
            scanstring('{"Missing colon" null}', 2, None, True),
            (u'Missing colon', 16))

        self.assertEqual(
            scanstring('{"Double colon":: null}', 2, None, True),
            (u'Double colon', 15))

        self.assertEqual(
            scanstring('{"Comma instead of colon", null}', 2, None, True),
            (u'Comma instead of colon', 25))

        self.assertEqual(
            scanstring('["Colon instead of comma": false]', 2, None, True),
            (u'Colon instead of comma', 25))

        self.assertEqual(
            scanstring('["Bad value", truth]', 2, None, True),
            (u'Bad value', 12))

    def test_surrogates(self):
        scanstring = self.json.decoder.scanstring
        def assertScan(given, expect):
            self.assertEqual(scanstring(given, 1, None, True),
                             (expect, len(given)))
            # XXX for python2 json compatibility: no unicode is allowed as input json-string
            # if not isinstance(given, unicode):
            #     given = unicode(given)
            #     self.assertEqual(scanstring(given, 1, None, True),
            #                      (expect, len(given)))

        # surrogates = unichr(0xd834) + unichr(0xdd20)
        assertScan('"z\\ud834\\u0079x"', 'z\xed\xa0\xb4yx')
        assertScan('"z\\ud834\\udd20x"', 'z\xed\xa0\xb4\xed\xb4\xa0x')
        assertScan('"z\\ud834\\ud834\\udd20x"', 'z\xed\xa0\xb4\xed\xa0\xb4\xed\xb4\xa0x')
        assertScan('"z\\ud834x"', 'z\xed\xa0\xb4x')
        assertScan('"z\\ud834\udd20x12345"', 'z\xed\xa0\xb4\xed\xb4\xa0x12345')
        assertScan('"z\\udd20x"', 'z\xed\xb4\xa0x')
        assertScan('"z\ud834\udd20x"', 'z\xed\xa0\xb4\xed\xb4\xa0x')
        assertScan('"z\ud834\\udd20x"', 'z\xed\xa0\xb4\xed\xb4\xa0x')
        assertScan('"z\ud834x"', 'z\xed\xa0\xb4x')

    def test_bad_escapes(self):
        scanstring = self.json.decoder.scanstring
        bad_escapes = [
            '"\\"',
            '"\\x"',
            '"\\u"',
            '"\\u0"',
            '"\\u01"',
            '"\\u012"',
            '"\\uz012"',
            '"\\u0z12"',
            '"\\u01z2"',
            '"\\u012z"',
            '"\\u0x12"',
            '"\\u0X12"',
            '"\\ud834\\"',
            '"\\ud834\\u"',
            '"\\ud834\\ud"',
            '"\\ud834\\udd"',
            '"\\ud834\\udd2"',
            '"\\ud834\\uzdd2"',
            '"\\ud834\\udzd2"',
            '"\\ud834\\uddz2"',
            '"\\ud834\\udd2z"',
            '"\\ud834\\u0x20"',
            '"\\ud834\\u0X20"',
        ]
        for s in bad_escapes:
            with self.assertRaises(ValueError):
                # XXX for python2 json compatibility: py_scanstring does not have argument encoding in py3
                scanstring(s, 1, True)
                # scanstring(s, 1, None, True)

    def test_issue3623(self):
        self.assertRaises(ValueError, self.json.decoder.scanstring, b"xxx", 1,
                          "xxx")
        # XXX for python2 json compatibility: py3 does not have this issue
        # self.assertRaises(UnicodeDecodeError,
        #                   self.json.encoder.encode_basestring_ascii, b"xx\xff")

    def test_overflow(self):
        with self.assertRaises(OverflowError):
            self.json.decoder.scanstring(b"xxx", sys.maxsize+1)


class TestPyScanstring(TestScanstring, PyTest): pass



from .proc import (
    ProcError,
    command,
    command_ex,
    shell_script,
    start_process,
)

__all__ = [
    'ProcError',
    'command',
    'command_ex',
    'shell_script',
    'start_process',
]



#!/usr/bin/env python2
# coding: utf-8

import unittest

from pykit import mime
from pykit import ututil

dd = ututil.dd


class TestMime(unittest.TestCase):

    def test_get_by_filename(self):
        cases = (
            ('', 'application/octet-stream'),
            ('123', 'application/octet-stream'),
            ('file.123', 'application/vnd.lotus-1-2-3'),
            ('file.123.not_exist_suffix_aa', 'application/octet-stream'),
            ('file.json', 'application/json'),
            ('file.not_exist_suffix_aa', 'application/octet-stream'),
        )

        for inp, expected in cases:
            dd('inp, expected:', inp, ' ', expected)
            rst = mime.get_by_filename(inp)

            dd('rst:', rst)

            self.assertEqual(expected, rst)



#!/usr/bin/env python
# coding: utf-8

from collections import namedtuple

from .idbase import IDBase

BlockGroupIDLen = 16


class BlockGroupIDError(Exception):
    pass


class BlockGroupID(namedtuple('_BlockGroupID', 'block_size seq'), IDBase):

    _tostr_fmt = 'g{block_size:0>5}{seq:0>10}'

    @classmethod
    def parse(cls, block_group_id):
        if len(block_group_id) != BlockGroupIDLen:
            raise BlockGroupIDError('Block group id length should be {0}, but is {1}: {2}'.format(
                BlockGroupIDLen, len(block_group_id), block_group_id))

        size = int(block_group_id[1:6])
        seq = int(block_group_id[6:])

        return BlockGroupID(size, seq)

    def __new__(clz, block_size, seq):
        block_size = int(block_size)
        seq = int(seq)

        return super(BlockGroupID, clz).__new__(clz, block_size, seq)



from .redisutil import (
    get_client,
    wait_serve,
    normalize_ip_port,

    RedisChannel,
)

from .redis_proxy_cli import(
    KeyNotFoundError,
    RedisProxyError,
    SendRequestError,
    ServerResponseError,

    RedisProxyClient,
)

__all__ = [
    'get_client',
    'wait_serve',
    'normalize_ip_port',

    'RedisChannel',

    'KeyNotFoundError',
    'RedisProxyError',
    'SendRequestError',
    'ServerResponseError',

    'RedisProxyClient',
]



#!/usr/bin/env python2
# coding: utf-8

import logging
import socket

from pykit import awssign
from pykit import config
from pykit import http
from pykit import utfjson

logger = logging.getLogger(__name__)
DEFAULT_TIMEOUT = 2


class RedisProxyError(Exception):
    pass


class SendRequestError(RedisProxyError):
    pass


class KeyNotFoundError(RedisProxyError):
    pass


class ServerResponseError(RedisProxyError):
    pass


def _retry(func):

    def _wrapper(self, *args):
        retry_cnt = args[-1] or 0

        for _ in range(retry_cnt + 1):
            for ip, port in self.hosts:
                self.ip = ip
                self.port = port

                try:
                    return func(self, *args)

                except (http.HttpError, socket.error) as e:
                    logger.error('{e} while send request to redis proxy with {ip}:{p}'.format(
                                 e=repr(e), ip=ip, p=port))

        else:
            raise SendRequestError(repr(e))

    return _wrapper


class SetAPI(object):

    def __init__(self, cli, redis_op, mtd_info):
        self.cli = cli
        self.redis_op = redis_op.upper()
        self.http_mtd = mtd_info[0]
        self.args_count = mtd_info[1]
        self.opts = list(mtd_info[2]) + ['retry']

    def api(self, *args, **argkv):
        mtd_args = []

        for idx in range(self.args_count):
            if idx < len(args):
                mtd_args.append(args[idx])

            else:
                left_cnt = self.args_count - len(mtd_args)
                opt_name = self.opts[-left_cnt]
                mtd_args.append(argkv.get(opt_name, None))

        retry = mtd_args.pop()

        qs = {}
        # retry in opts, but it is not in qs
        qs_keys = list(self.opts[:-1])
        while len(qs_keys) > 0:
            qs[qs_keys.pop()] = mtd_args.pop()

        body = None
        if self.http_mtd == 'PUT':
            body = utfjson.dump(mtd_args.pop())

        path = [self.redis_op] + mtd_args

        return self.cli._api(self.http_mtd, path, body, qs, retry)


class RedisProxyClient(object):

    # http method, count of args, optional args name
    methods = {
        # get(key, retry=0)
        'get': ('GET', 2, ()),

        # set(key, val, expire=None, retry=0)
        'set': ('PUT', 4, ('expire',)),

        # hget(hashname, hashkey, retry=0)
        'hget': ('GET', 3, ()),

        # hset(hashname, hashkey, val, expire=None, retry=0)
        'hset': ('PUT', 5, ('expire',)),
    }

    def __init__(self, hosts, nwr=None, ak_sk=None, timeout=None):
        self.hosts = hosts

        if nwr is None:
            nwr = config.rp_cli_nwr

        if ak_sk is None:
            ak_sk = config.rp_cli_ak_sk

        self.n, self.w, self.r = nwr
        self.access_key, self.secret_key = ak_sk

        self.timeout = timeout or DEFAULT_TIMEOUT
        self.ver = '/redisproxy/v1'

        for mtd_name, mtd_info in self.methods.items():
            api_obj = SetAPI(self, mtd_name, mtd_info)
            setattr(self, mtd_name, api_obj.api)

    def _sign_req(self, req):
        sign_payload = True if 'body' in req else False
        signer = awssign.Signer(self.access_key, self.secret_key)
        sign_ctx = signer.add_auth(req, query_auth=True, sign_payload=sign_payload)
        logger.info('signing details: {ctx}'.format(ctx=sign_ctx))

    def _make_req_uri(self, params, qs):
        path = [self.ver]
        path.extend(params)

        qs_list = [
            'n={n}'.format(n=self.n),
            'w={w}'.format(w=self.w),
            'r={r}'.format(r=self.r),
        ]
        for k, v in qs.items():
            if v is None:
                continue

            qs_list.append('{k}={v}'.format(k=k, v=v))

        return '{p}?{qs}'.format(p='/'.join(path), qs='&'.join(qs_list))

    def _req(self, req):
        if 'headers' not in req:
            req['headers'] = {
                'host': '{ip}:{port}'.format(ip=self.ip, port=self.port),
            }

        elif 'host' not in req['headers']:
            req['headers']['host'] = '{ip}:{port}'.format(ip=self.ip, port=self.port)

        body = req.get('body', None)
        if body is not None:
            req['headers']['Content-Length'] = len(body)

        self._sign_req(req)

        cli = http.Client(self.ip, self.port, self.timeout)
        cli.send_request(req['uri'], method=req['verb'], headers=req['headers'])
        if body is not None:
            cli.send_body(req['body'])

        cli.read_response()
        res = cli.read_body(None)

        msg = 'Status:{s} req:{req} res:{res} server:{ip}:{p}'.format(
              s=cli.status, req=repr(req), res=repr(res),
              ip=self.ip, p=self.port)

        if cli.status == 404:
            raise KeyNotFoundError(msg)

        elif cli.status != 200:
            raise ServerResponseError(msg)

        return res

    @_retry
    def _api(self, verb, path, body, qs, retry):
        req = {
            'verb': verb,
            'uri': self._make_req_uri(path, qs),
        }

        if body is not None:
            req['body'] = body

        if verb == 'GET':
            rst = self._req(req)
            return utfjson.load(rst)

        else:
            self._req(req)



#!/usr/bin/env python
# coding: utf-8

from pykit import utfjson


class RedisKeyValue(object):

    def __init__(self, redis_cli, get_path=None, load=None):
        self.cli = redis_cli
        self.get_path = get_path
        self.load = load

    def _get_path(self, key):
        if self.get_path is not None:
            return self.get_path(key)

        return key

    def set(self, key, value):
        key = self._get_path(key)
        value = utfjson.dump(value)
        self.cli.set(key, value)

    def get(self, key):
        key = self._get_path(key)
        val = self.cli.get(key)

        if self.load is not None:
            return self.load(val)

        return utfjson.load(val)

    def hget(self, hashname, hashkey):
        hashkey = self._get_path(hashkey)
        val = self.cli.hget(hashname, hashkey)

        if self.load is not None:
            return self.load(val)

        return utfjson.load(val)

    def hset(self, hashname, hashkey, value):
        hashkey = self._get_path(hashkey)
        value = utfjson.dump(value)
        self.cli.hset(hashname, hashkey, value)


class RedisValue(RedisKeyValue):

    def _get_path(self, key):
        return self.get_path()

    def set(self, value):
        return super(RedisValue, self).set('', value)

    def get(self):
        return super(RedisValue, self).get('')



#!/usr/bin/env python2
# coding: utf-8

import socket
import subprocess32
import time
import unittest

import docker
import MySQLdb

from pykit import mysqlconnpool
from pykit import ututil
from pykit import utdocker

dd = ututil.dd

mysql_test_ip = '192.168.52.40'
mysql_test_password = '123qwe'
mysql_test_port = 3306
mysql_test_user = 'root'
mysql_test_name = 'mysql_test'
mysql_test_tag = 'daocloud.io/mysql:5.7.13'


class ShellError(Exception):
    pass


class Testmysqlconnpool(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        utdocker.pull_image(mysql_test_tag)

    def setUp(self):

        utdocker.start_container(
                mysql_test_name,
                mysql_test_tag,
                ip=mysql_test_ip,
                env={
                    'MYSQL_ROOT_PASSWORD': mysql_test_password,
                },
                port_bindings={
                    mysql_test_port: mysql_test_port
                }

        )

        addr = (mysql_test_ip, mysql_test_port)

        # some time it takes several seconds to start listening
        for ii in range(40):
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.connect(addr)
                break
            except socket.error:
                dd('trying to connect to {0} failed'.format(str(addr)))
                sock.close()
                time.sleep(.4)
        else:
            raise

        self.pool = mysqlconnpool.make({
            'host': mysql_test_ip,
            'port': mysql_test_port,
            'user': mysql_test_user,
            'passwd': mysql_test_password,
        })

    def tearDown(self):
        utdocker.remove_container(mysql_test_name)

    def test_pool_name(self):

        pool = self.pool

        name = '{0}:{1}'.format(mysql_test_ip, mysql_test_port)

        self.assertEqual(name, pool('stat')['name'])

    def test_initial_stat(self):

        pool = self.pool

        dd('pool stat: {0}'.format(pool('stat')))

        self.assertEqual(0, pool('stat')['create'])
        self.assertEqual(0, pool('stat')['pool_get'])
        self.assertEqual(0, pool('stat')['pool_put'])

    def test_get_connection(self):

        pool = self.pool

        with pool() as conn:

            dd(conn)
            dd('pool stat: {0}'.format(pool('stat')))

            self.assertEqual(0, pool('stat')['pool_put'])

        dd('pool stat: {0}'.format(pool('stat')))

        self.assertEqual(1, pool('stat')['create'])
        self.assertEqual(0, pool('stat')['pool_get'])
        self.assertEqual(1, pool('stat')['pool_put'])

    def test_connection_is_working(self):

        pool = self.pool

        with pool() as conn:

            dd(conn)

            conn.query('show databases')
            rst = conn.store_result()
            rst = rst.fetch_row(rst.num_rows())

            databases = [x[0] for x in rst]
            self.assertTrue('mysql' in databases)

    def test_get_connection_twice(self):

        pool = self.pool

        c0 = None
        with pool() as conn:
            c0 = conn

        # the second time do not create new connection
        with pool() as conn:

            self.assertTrue(c0 is conn)

            dd('pool stat: {0}'.format(pool('stat')))

            self.assertEqual(1, pool('stat')['create'])
            self.assertEqual(1, pool('stat')['pool_get'])

        # after put back
        dd('pool stat: {0}'.format(pool('stat')))

        self.assertEqual(2, pool('stat')['pool_put'])

    def test_raise_in_with(self):

        pool = self.pool

        try:
            with pool() as conn:
                dd(conn)
                raise ValueError(1)
        except ValueError:
            pass

        dd('pool stat: {0}'.format(pool('stat')))

        self.assertEqual(1, pool('stat')['create'])
        self.assertEqual(0, pool('stat')['pool_get'])
        self.assertEqual(0, pool('stat')['pool_put'])

    def test_conn_query(self):

        pool = self.pool

        with pool() as conn:
            rst = mysqlconnpool.conn_query(conn, 'show databases')

        dd('databases: {0}'.format(rst))

        databases = [x['Database'] for x in rst]
        self.assertTrue('mysql' in databases)

    def test_query(self):

        pool = self.pool

        rst = pool.query('show databases')

        dd('databases: {0}'.format(rst))

        databases = [x['Database'] for x in rst]
        self.assertTrue('mysql' in databases)

    def test_query_retry(self):

        pool = self.pool

        sql = (
            'set session wait_timeout=1;'
        )
        pool.query(sql)
        pool.query('show variables like "%timeout%";')

        with pool() as conn:
            time.sleep(2)
            with self.assertRaises(MySQLdb.OperationalError):
                print conn.query('show databases')

        # no error raise from above, thus a timed out conn has been left in
        # pool
        stat = pool('stat')
        dd('stat after timeout', stat)
        self.assertEqual(1, stat['create'], 'created 1 conn')

        # use previous conn, timed out and retry.
        pool.query('show databases', retry=1)

        stat = pool('stat')
        dd('stat after retry', stat)
        self.assertEqual(2, stat['create'], 'created another conn for retry')

    def test_query_array_result(self):

        pool = self.pool

        rst = pool.query('show databases', use_dict=False)

        dd('databases: {0}'.format(rst))

        databases = [x[0] for x in rst]
        self.assertTrue('mysql' in databases)

    def test_query_charset(self):

        sql_create_table = ('CREATE TABLE `test`.`q` ( `name` varchar(128) NOT NULL)'
                            ' ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_bin')

        pool = mysqlconnpool.make({
            'host': mysql_test_ip,
            'port': mysql_test_port,
            'user': mysql_test_user,
            'passwd': mysql_test_password,
            'charset': 'utf8',
        })

        rst = pool.query('CREATE DATABASE `test`')
        dd('create database rst: ', rst)

        rst = pool.query(sql_create_table)
        dd('create table rst: ', rst)

        rst = pool.query('INSERT INTO `test`.`q` (`name`) VALUES ("我")')
        dd('insert rst: ', rst)

        rst = pool.query('SELECT * FROM `test`.`q`')
        dd('select rst: ', rst)

        self.assertEqual(({'name': u'\u6211'},), rst)



#!/usr/bin/env python
# coding: utf-8

from kazoo.client import KazooClient

from pykit import config
from pykit import utfjson

from . import zkutil


class ZKConf(object):

    """
    zk dir:
        <prefix>/record/<key>
        <prefix>/lock/<key>
        <prefix>/tx/
                    alive/0000000001
                    journal/0000000001
                    state/0000000001
                    txidset
                    txid_maker
        <prefix>/seq/<key>

    alive:      Contains `ephemeral` node each of which represents a transaction.
                Its modifications is used.

    journal:    Contains journal transaction modifications. Each of them is a complete transaction.

    txidset: Last committed txid number.
    """

    def __init__(self,
                 hosts=None,
                 tx_dir=None,
                 seq_dir=None,
                 record_dir=None,
                 lock_dir=None,
                 node_id=None,
                 auth=None,
                 acl=None
                 ):

        self.conf = {
            'hosts':      hosts,
            'tx_dir':     tx_dir,
            'seq_dir':    seq_dir,
            'record_dir': record_dir,
            'lock_dir':   lock_dir,
            'node_id':    node_id,
            'auth':       auth,
            'acl':        acl,
        }

    def hosts(self): return self._get_config('hosts')

    def record_dir(self): return self._get_config('record_dir')

    def lock_dir(self): return self._get_config('lock_dir')

    def node_id(self): return self._get_config('node_id')

    def auth(self): return self._get_config('auth')

    def acl(self): return self._get_config('acl')

    def lock(self, key=''): return ''.join([self.lock_dir(), _dump_txid(key)])

    def record(self, key=''): return ''.join([self.record_dir(), key])

    def seq_dir(self): return self._get_config('seq_dir')

    def seq(self, key=''): return ''.join([self.seq_dir(), key])

    def tx_dir(self): return self._get_config('tx_dir')

    def tx_alive(self, txid=''): return ''.join([self.tx_dir(), 'alive/', _dump_txid(txid)])

    def tx_state(self, txid=''): return ''.join([self.tx_dir(), 'state/', _dump_txid(txid)])

    def journal(self, txid=''): return ''.join([self.tx_dir(), 'journal/', _dump_txid(txid)])

    def txidset(self): return ''.join([self.tx_dir(), 'txidset'])

    def txid_maker(self): return ''.join([self.tx_dir(), 'txid_maker'])

    def kazoo_digest_acl(self):
        a = self.acl()
        if a is None:
            return a

        return zkutil.make_kazoo_digest_acl(a)

    def kazoo_auth(self):

        a = self.auth()
        if a is None:
            return None

        return a[0], a[1] + ':' + a[2]

    def _get_config(self, name):

        if self.conf[name] is None:
            return getattr(config, 'zk_' + name)
        else:
            return self.conf[name]


class KazooClientExt(KazooClient):

    def __init__(self, zkclient, json=True):

        if isinstance(zkclient, KazooClientExt):
            self._zk = zkclient._zk
            self._zkconf = ZKConf(**zkclient._zkconf.conf)

        elif isinstance(zkclient, KazooClient):
            self._zk = zkclient
            self._zkconf = None
        else:
            raise TypeError('invalid zkclient type: expect KazooClient or KazooClientExt')

        self._json = json

    def __getattr__(self, n):
        return getattr(self._zk, n)

    def _jl(self, v):
        if self._json:
            return utfjson.load(v)
        else:
            return v

    def _jd(self, v):
        if self._json:
            return utfjson.dump(v)
        else:
            return v

    def get(self, path, watch=None):
        val, zstat = self._zk.get(path, watch=watch)
        return self._jl(val), zstat

    def set(self, path, value, version=-1):
        value = self._jd(value)
        return self._zk.set(path, value, version=version)

    def create(self, path, value=b"", acl=None, ephemeral=False,
               sequence=False, makepath=False):

        value = self._jd(value)

        return self._zk.create(path, value=value,
                               acl=acl,
                               ephemeral=ephemeral,
                               sequence=sequence,
                               makepath=makepath)


def _dump_txid(txid):
    if isinstance(txid, str):
        return txid
    elif isinstance(txid, (int, long)):
        return '%010d' % txid
    else:
        raise TypeError('invalid type txid: ' + repr(txid))


def kazoo_client_ext(zk, json=True):
    """
    return zkclient created or original zkclient, and if zkclient is created
    """

    zkconf = None

    if isinstance(zk, str):
        zkconf = ZKConf(hosts=zk)

    if isinstance(zk, dict):
        zkconf = ZKConf(**zk)

    if isinstance(zk, ZKConf):
        zkconf = zk

    if zkconf is None:
        zkconf = ZKConf()
        owning = False
    else:
        zk = KazooClient(zkconf.hosts())
        owning = True

    zkclient = KazooClientExt(zk, json=json)
    zkclient._zkconf = zkconf

    zkclient.start()

    auth = zkconf.kazoo_auth()
    if auth is not None:
        zkclient.add_auth(*auth)

    return zkclient, owning



#!/usr/bin/env python2
# coding: utf-8

import time
import unittest

from pykit import ratelimiter


class TestRateLimiter(unittest.TestCase):
    def test_set_token_per_second(self):
        r = ratelimiter.RateLimiter(1, 2)

        self.assertEqual(1, r.token_per_second)
        r.set_token_per_second(10)

        self.assertEqual(10, r.token_per_second)

        r.set_token_per_second(5)
        self.assertEqual(5, r.token_per_second)

    def test_set_capacity(self):
        r = ratelimiter.RateLimiter(2, 10)
        r.set_capacity(100)
        self.assertEqual(100, r.capacity)

        r.set_capacity(1)
        self.assertEqual(1, r.capacity)
        self.assertAlmostEqual(1, r.get_stored(), places=1)

    def test_consume_and_get_stored(self):
        r = ratelimiter.RateLimiter(1, 0.4)
        r.consume(0.4)
        self.assertAlmostEqual(0, r.get_stored(), places=1)
        time.sleep(0.5)
        self.assertAlmostEqual(0.4, r.get_stored(), places=1)
        r.consume(1)
        self.assertAlmostEqual(-0.6, r.get_stored(), places=1)
        time.sleep(0.01)
        self.assertAlmostEqual(-0.59, r.get_stored(), places=1)

    def test_consume_and_get_stored_token_time(self):
        r = ratelimiter.RateLimiter(10, 100)

        consume_time = time.time() + 1
        r.consume(100, consume_time)
        self.assertAlmostEqual(-80, r.get_stored(consume_time), places=1)
        self.assertEqual(r.get_stored(), r.get_stored(consume_time - 1))

        self.assertAlmostEqual(-70, r.get_stored(consume_time + 1), places=1)



import Queue
import threading

from pykit import heap

default_priority = 10.0

Empty = Queue.Empty


class Producer(object):

    def __init__(self, producer_id, priority, iterable):
        self.consumed = 0
        self.iterable_lock = threading.RLock()
        self.stat = {'get': 0}
        self.cmp_key = (0, 0)

        self.producer_id = producer_id
        self.set_priority(priority)
        self.set_iterable(iterable)

    def get(self):

        with self.iterable_lock:
            try:
                val = self.iterable.next()
                self.stat['get'] += 1
                self.consume()
                return val
            except StopIteration:
                raise Empty('no more item in ' + str(self))

    def set_priority(self, priority):

        priority = float(priority)

        if priority <= 0:
            raise ValueError('priority can not be less or euqal 0: ' + str(priority))

        self.priority = priority
        self.item_cost = default_priority / float(self.priority)
        self.cmp_key = (self.consumed, self.item_cost)

    def set_iterable(self, iterable):
        self.iterable = iter(iterable)

    def consume(self):
        self.consumed += self.item_cost
        self.cmp_key = (self.consumed, self.item_cost)

    def __str__(self):
        return '[{producer_id}={priority} c={consumed}]'.format(
            producer_id=self.producer_id,
            priority=self.priority,
            consumed=self.consumed,
        )

    def __lt__(self, b):
        return self.cmp_key < b.cmp_key


class PriorityQueue(object):

    def __init__(self):
        self.producer_by_id = {}

        # empty_heap: stores all empty Producer.
        #             Empty produer means it has raised an Empty exception when
        #             calling Producer.get().
        #             `Empty` exception raised means it has no more item to
        #             produce.
        #
        # consumable_heap: stores all non-empty Producer, less consumed
        #             Producer is at high position in heap.
        self.empty_heap = heap.RefHeap()
        self.consumable_heap = heap.RefHeap()

        self.heap_lock = threading.RLock()

    def add_producer(self, producer_id, priority, iterable):

        with self.heap_lock:
            if producer_id not in self.producer_by_id:
                p = Producer(producer_id, priority, iterable)
                self.producer_by_id[producer_id] = p
            else:
                # if exist, update its priority and iterable.
                p = self.producer_by_id[producer_id]
                p.set_priority(priority)
                p.set_iterable(iterable)

            # Every time add a (may be existent) queue, treat it as consumable
            self._remove_from_heaps(p)
            self.consumable_heap.push(p)

    def remove_producer(self, producer_id, ignore_not_found=False):

        with self.heap_lock:
            if producer_id not in self.producer_by_id and ignore_not_found:
                return

            p = self.producer_by_id[producer_id]
            self._remove_from_heaps(p)

            del self.producer_by_id[producer_id]

    def _remove_from_heaps(self, producer):

        try:
            self.empty_heap.remove(producer)
        except heap.NotFound:
            pass

        try:
            self.consumable_heap.remove(producer)
        except heap.NotFound:
            pass

    def get(self):

        while True:

            with self.heap_lock:
                try:
                    p = self.consumable_heap.get()
                except heap.Empty:
                    raise Empty('no more queue has any item')

                try:
                    # NOTE: if p.iterable blocks, everything is blocked
                    val = p.get()

                except Empty:

                    self.consumable_heap.remove(p)
                    self.empty_heap.push(p)

                    # try next consumable queue
                    continue

                self.consumable_heap.sift(p)

                return val

    def __str__(self):
        qs = []
        for cq in self.producer_by_id.values():
            qs.append(str(cq))

        return ' '.join(qs)



#!/usr/bin/env python
# coding: utf-8


import copy
import logging

logger = logging.getLogger(__name__)


class CASError(Exception):
    pass


class CASConflict(CASError):
    pass


class CASRecord(object):
    def __init__(self, v, stat, n):
        self.v = v
        self.stat = stat
        self.n = n


def cas_loop(getter, setter, args=(), kwargs=None, conflicterror=CASConflict):

    if kwargs is None:
        kwargs = {}

    i = -1
    while True:

        i += 1

        val, stat = getter(*args, **kwargs)
        rec = CASRecord(copy.deepcopy(val), stat, i)

        yield rec

        if rec.v == val:
            # nothing to update
            return

        try:
            setter(*(args + (rec.v, rec.stat)), **kwargs)
            return
        except conflicterror as e:
            logger.info(repr(e) + ' while cas set')
            continue



#!/usr/bin/env python2
# coding: utf-8

import threading
import time


class RateLimiter(object):
    def __init__(self, token_per_second, capacity):
        self.token_per_second = token_per_second
        self.capacity = capacity
        self.stored = float(min(token_per_second, capacity))
        self.sync_time = time.time()

        self.lock = threading.RLock()

    def consume(self, consumed, token_time=None):
        with self.lock:
            self._sync(token_time=token_time)
            self.stored = self.stored - consumed

    def _sync(self, token_time=None):
        with self.lock:
            new_sync_time = token_time or time.time()

            if new_sync_time <= self.sync_time:
                return self.stored

            new_tokens = (new_sync_time - self.sync_time) * self.token_per_second
            self.stored = min(self.capacity, self.stored + new_tokens)
            self.sync_time = new_sync_time

            return self.stored

    def set_token_per_second(self, token_per_second):
        self.token_per_second = token_per_second

    def set_capacity(self, capacity):
        with self.lock:
            self._sync()
            self.capacity = capacity
            self.stored = min(self.stored, self.capacity)

    def get_stored(self, token_time=None):
        return self._sync(token_time=token_time)



import textwrap
from StringIO import StringIO
from pykit.p3json.test import PyTest


class TestIndent(object):
    def test_indent(self):
        h = [['blorpie'], ['whoops'], [], 'd-shtaeou', 'd-nthiouh', 'i-vhbjkhnth',
             {'nifty': 87}, {'field': 'yes', 'morefield': False} ]

        expect = textwrap.dedent("""\
        [
          [
            "blorpie"
          ],
          [
            "whoops"
          ],
          [],
          "d-shtaeou",
          "d-nthiouh",
          "i-vhbjkhnth",
          {
            "nifty": 87
          },
          {
            "field": "yes",
            "morefield": false
          }
        ]""")


        d1 = self.dumps(h)
        d2 = self.dumps(h, indent=2, sort_keys=True, separators=(',', ': '))

        h1 = self.loads(d1)
        h2 = self.loads(d2)

        self.assertEqual(h1, h)
        self.assertEqual(h2, h)
        self.assertEqual(d2, expect)

    def test_indent0(self):
        h = {3: 1}
        def check(indent, expected):
            d1 = self.dumps(h, indent=indent)
            self.assertEqual(d1, expected)

            sio = StringIO()
            self.json.dump(h, sio, indent=indent)
            self.assertEqual(sio.getvalue(), expected)

        # indent=0 should emit newlines
        check(0, '{\n"3": 1\n}')
        # indent=None is more compact
        check(None, '{"3": 1}')


class TestPyIndent(TestIndent, PyTest): pass



import unittest

from pykit import ututil
from pykit.cachepool import CachePool
from pykit.cachepool import CachePoolGeneratorError
from pykit.cachepool import make_wrapper

dd = ututil.dd


# Exception that is used for test.
class ErrorInReuse(Exception):
    pass


class StillAllowToReuse(Exception):
    pass


class TestCachePool(unittest.TestCase):

    def test_initial_stat(self):

        pool = CachePool(
            generator,
        )

        dd('pool stat: {0}'.format(pool.stat))

        self.assertEqual(0, pool.stat['new'])
        self.assertEqual(0, pool.stat['put'])
        self.assertEqual(0, pool.stat['get'])

    def test_generator_is_not_callable(self):

        try:
            CachePool(
                1,
            )
        except CachePoolGeneratorError as e:
            self.assertEqual(str(e), 'generator is not callable')

        else:
            self.fail("didn't catch CachePoolGeneratorError")

    def test_generator_run_error_1(self):

        try:
            pool = CachePool(
                lambda x, y: (x, y),
            )
            pool.get()

        except TypeError as e:
            dd(repr(e))

        else:
            self.fail("run generator error")

    def test_generator_run_error_2(self):

        try:
            pool = CachePool(
                lambda x, y: (x, y),
                generator_args=[1, 2],
                generator_argkw={'k': 'v'},
            )
            pool.get()

        except TypeError as e:
            dd(repr(e))

        else:
            self.fail("run generator error")

    def test_generator_run_error_3(self):

        try:
            pool = CachePool(
                lambda x, y: (x, y),
                generator_argkw=[1, 2],
            )
            pool.get()

        except TypeError as e:
            dd(repr(e))

        else:
            self.fail("run generator error")

    def test_close_callback_error(self):

        try:
            pool = CachePool(
                generator,
                close_callback=close_callback_error,
            )
            element = pool.get()
            pool.close(element)

        except AttributeError as e:
            dd(repr(e))

        else:
            self.fail("run close_callback error")

    def test_close_callback(self):

        pool = CachePool(
            generator,
            close_callback=close_callback,
        )
        element = pool.get()

        self.assertFalse(element.closed)
        pool.close(element)
        self.assertTrue(element.closed)

    def test_close_callback_when_queue_full(self):

        pool = CachePool(
            generator,
            pool_size=1,
            close_callback=close_callback,
        )

        dd('pool qsize: {0}'.format(pool.queue.qsize()))
        element1 = pool.get()
        element2 = pool.get()

        pool.put(element1)
        pool.put(element2)

        dd('pool qsize: {0}'.format(pool.queue.qsize()))
        self.assertFalse(element1.closed)
        self.assertTrue(element2.closed)

    def test_pool_size(self):

        pool = CachePool(
            generator,
            pool_size=1,
            close_callback=close_callback,
        )

        element1 = pool.get()
        element2 = pool.get()

        self.assertEqual(0, pool.queue.qsize())
        pool.put(element1)
        self.assertEqual(1, pool.queue.qsize())
        pool.put(element2)
        self.assertEqual(1, pool.queue.qsize())

    def test_get_element(self):

        pool = CachePool(
            generator,
        )

        element = pool.get()

        dd('pool stat: {0}'.format(pool.stat))
        self.assertEqual(1, pool.stat['new'])
        self.assertEqual(1, pool.stat['get'])

        pool.put(element)

        dd('pool stat: {0}'.format(pool.stat))
        self.assertEqual(1, pool.stat['put'])

    def test_get_element_twice(self):

        pool = CachePool(
            generator,
        )

        element1 = pool.get()
        pool.put(element1)

        dd('pool stat: {0}'.format(pool.stat))
        self.assertEqual(1, pool.stat['new'])

        element2 = pool.get()
        pool.put(element2)

        dd('pool stat: {0}'.format(pool.stat))
        self.assertEqual(2, pool.stat['get'])
        self.assertIs(element1, element2)

    def test_args_argkw(self):

        args = [1, 'x']
        argkw = {'key': 'value'}

        pool = CachePool(
            generator,
            generator_args=args,
            generator_argkw=argkw,
        )

        element = pool.get()

        self.assertEqual(element.args[0], args[0])
        self.assertEqual(element.args[1], args[1])
        self.assertEqual(element.argkw, argkw)


class TestWrapper(unittest.TestCase):

    def test_new(self):

        pool = CachePool(
            generator,
        )
        wrapper = make_wrapper(
            pool,
        )

        element1 = element2 = None
        with wrapper() as ele:
            element1 = ele

            with wrapper() as ele:
                element2 = ele

        _element2 = pool.get()
        _element1 = pool.get()

        self.assertIsNot(element1, element2)
        self.assertIs(element1, _element1)
        self.assertIs(element2, _element2)

    def test_reuse(self):

        pool = CachePool(
            generator,
        )
        wrapper = make_wrapper(
            pool,
        )

        element1 = element2 = None
        with wrapper() as ele:
            element1 = ele

        with wrapper() as ele:
            element2 = ele

        _element1 = pool.get()
        _element2 = pool.get()

        self.assertIs(element1, element2)
        self.assertIs(element1, _element1)
        self.assertIs(element2, _element1)
        self.assertIsNot(element2, _element2)

    def test_reuse_decider_error(self):

        pool = CachePool(
            generator,
        )
        wrapper = make_wrapper(
            pool,
            reuse_decider=reuse_decider_error,
        )

        with wrapper() as ele:
            (ele)

        try:
            with wrapper() as ele:
                raise Exception()
        except ErrorInReuse:
            pass

        except:
            self.fail("didn't run reuse_decider")

        else:
            self.fail("didn't run reuse_decider")

    def test_reuse_decider_reuse(self):

        pool = CachePool(
            generator,
            close_callback=close_callback,
        )
        wrapper = make_wrapper(
            pool,
            reuse_decider=reuse_for_acceptable_errors,
        )

        element1 = element2 = None

        try:
            with wrapper() as ele:
                element1 = ele
                raise StillAllowToReuse()
        except:
            pass

        with wrapper() as ele:
            element2 = ele

        self.assertIs(element1, element2)
        self.assertFalse(element1.closed)

    def test_reuse_decider_drop(self):

        pool = CachePool(
            generator,
            close_callback=close_callback,
        )
        wrapper = make_wrapper(
            pool,
            reuse_decider=reuse_for_acceptable_errors,
        )

        element1 = element2 = None

        try:
            with wrapper() as ele:
                element1 = ele
                raise Exception()
        except:
            pass

        with wrapper() as ele:
            element2 = ele

        self.assertIsNot(element1, element2)
        self.assertTrue(element1.closed)
        self.assertFalse(element2.closed)


class Element(object):

    def __init__(self, *args, **argkw):
        self.args = args
        self.argkw = argkw

        self.closed = False

    def close(self):
        self.closed = True

    def do(self):
        print self.args
        print self.argkw


def generator(*args, **argkw):
    return Element(*args, **argkw)


def close_callback(element):
    element.close()


def close_callback_error(element):
    element._close()


def reuse_for_acceptable_errors(errtype, errval, _traceback):
    return errtype in (StillAllowToReuse, )


def reuse_decider_error(errtype, errval, _traceback):
    raise ErrorInReuse()



from .cachepool import (

    make_wrapper,

    CachePool,
    CacheWrapper,

    CachePoolError,
    CachePoolGeneratorError,
)

__all__ = [
    "make_wrapper",

    "CachePool",
    "CacheWrapper",

    "CachePoolError",
    "CachePoolGeneratorError",
]



#!/usr/bin/env python
# coding: utf-8

import bisect
import copy

from pykit import rangeset


class BlockNotInRegion(Exception):
    pass


class Region(dict):

    fields_default = {'idc': None,
                      'range': [None, None],
                      'levels': [], }

    def __init__(self, region=None, **argkv):

        self.update(copy.deepcopy(self.fields_default))

        if region is not None:
            self.update(copy.deepcopy(region))

        self.update(copy.deepcopy(argkv))

        self['range'] = rangeset.Range(*self['range'])

        for level, blocks in enumerate(self['levels']):
            self['levels'][level] = rangeset.RangeDict(blocks)

    def need_merge(self, source, targets):

        src_blk_size = source[2]['size']

        target_blk_size = 0
        for target in targets:
            target_blk_size += target[2]['size']

        return 4 * src_blk_size >= target_blk_size

    def find_moved_to_level(self, src_block, src_level, region_levels):

        for level in reversed(range(src_level)):

            blocks = region_levels[level]
            overlapped = blocks.find_overlapped(src_block)

            if len(overlapped) != 0:
                level += 1
                return level

        return 0

    def move_down(self):

        region_levels = self['levels']
        moved_blocks = []

        for level, src_blocks in enumerate(region_levels):
            if level == 0:
                continue

            for src in src_blocks[:]:
                target_level = self.find_moved_to_level(
                    src, level, region_levels)
                if level == target_level:
                    continue

                moved_blocks.append((level, target_level, src))

                region_levels[level].remove(src)
                region_levels[target_level].add(src[:2], src[2])

        while region_levels[-1] == []:
            region_levels.pop(-1)

        return moved_blocks

    def find_merge(self):

        region_levels = self['levels']

        for level, src_blocks in enumerate(region_levels):
            if level == 0:
                continue

            lower_blocks = region_levels[level - 1]

            for src in src_blocks:
                overlapped = lower_blocks.find_overlapped(src)

                if len(overlapped) == 0:
                    continue

                if self.need_merge(src, overlapped):
                    return (level, src, overlapped)

    def list_block_ids(self, start_block_id=None):

        block_ids = []

        for blocks in self['levels']:
            level_bids = [b[2]['block_id'] for b in blocks]
            block_ids.extend(level_bids)

        block_ids.sort()

        if start_block_id is not None:
            idx = bisect.bisect_left(block_ids, start_block_id)
            block_ids = block_ids[idx:]

        return block_ids

    def replace_block_id(self, block_id, new_block_id):

        for blocks in self['levels']:

            for block in blocks:
                if block[2]['block_id'] == block_id:
                    block[2]['block_id'] = new_block_id
                    return

        raise BlockNotInRegion('block_id: %s' % block_id)



#!/usr/bin/env python2
# coding: utf-8

import binascii
import errno
import hashlib
import os
import sys
import time

import psutil

from pykit import config
from pykit import timeutil

READ_BLOCK = 32 * 1024 * 1024
WRITE_BLOCK = 32 * 1024 * 1024


class FSUtilError(Exception):
    pass


class NotMountPoint(FSUtilError):
    pass


def assert_mountpoint(path):
    if not os.path.ismount(path):
        raise NotMountPoint(path)


def get_all_mountpoint(all=False):
    partitions = psutil.disk_partitions(all=all)
    prt_by_mp = [x.mountpoint for x in partitions]
    return prt_by_mp


def get_mountpoint(path):

    path = os.path.realpath(path)

    prt_by_mountpoint = get_disk_partitions()

    while path != '/' and path not in prt_by_mountpoint:
        path = os.path.dirname(path)

    return path


def get_device(path):

    prt_by_mountpoint = get_disk_partitions()

    mp = get_mountpoint(path)

    return prt_by_mountpoint[mp]['device']


def get_device_fs(device):

    prt_by_mp = get_disk_partitions()

    for prt in prt_by_mp.values():
        if device == prt['device']:
            return prt['fstype']
    else:
        return 'unknown'


def get_disk_partitions(all=True):

    partitions = psutil.disk_partitions(all=all)

    by_mount_point = {}
    for pt in partitions:
        # OrderedDict([
        #      ('device', '/dev/disk1'),
        #      ('mountpoint', '/'),
        #      ('fstype', 'hfs'),
        #      ('opts', 'rw,local,rootfs,dovolfs,journaled,multilabel')])
        by_mount_point[pt.mountpoint] = _to_dict(pt)

    return by_mount_point


def get_path_fs(path):

    mp = get_mountpoint(path)
    prt_by_mp = get_disk_partitions()

    return prt_by_mp[mp]['fstype']


def get_path_usage(path):

    space_st = os.statvfs(path)

    # f_bavail: without blocks reserved for super users
    # f_bfree:  with    blocks reserved for super users
    avail = space_st.f_frsize * space_st.f_bavail

    capa = space_st.f_frsize * space_st.f_blocks
    used = capa - avail

    return {
        'total': capa,
        'used': used,
        'available': avail,
        'percent': float(used) / capa,
    }


def get_path_inode_usage(path):

    inode_st = os.statvfs(path)

    available = inode_st.f_favail
    total = inode_st.f_files
    used = total - available

    return {
        'total': total,
        'used': used,
        'available': available,
        'percent': float(used) / total,
    }


def makedirs(*paths, **kwargs):
    mode = kwargs.get('mode', 0755)
    uid = kwargs.get('uid') or config.uid
    gid = kwargs.get('gid') or config.gid

    path = os.path.join(*paths)

    # retry to deal with concurrent check-and-then-set issue
    for ii in range(2):

        if os.path.isdir(path):

            if uid is not None and gid is not None:
                os.chown(path, uid, gid)

            return

        try:
            os.makedirs(path, mode=mode)
            if uid is not None and gid is not None:
                os.chown(path, uid, gid)
        except OSError as e:
            if e.errno == errno.EEXIST:
                # concurrent if-exist and makedirs
                pass
            else:
                raise
    else:
        raise


def get_sub_dirs(path):
    files = os.listdir(path)

    sub_dirs = []
    for f in files:
        if os.path.isdir(os.path.join(path, f)):
            sub_dirs.append(f)

    sub_dirs.sort()

    return sub_dirs


def read_file(path):
    with open(path, 'r') as f:
        return f.read()


def write_file(path, fcont, uid=None, gid=None, atomic=False, fsync=True):

    if not atomic:
        return _write_file(path, fcont, uid, gid, fsync)

    tmp_path = '{path}._tmp_.{pid}_{timestamp}'.format(
        path=path,
        pid=os.getpid(),
        timestamp=timeutil.ns(),
    )
    _write_file(tmp_path, fcont, uid, gid, fsync)

    try:
        os.rename(tmp_path, path)
    except EnvironmentError:
        os.remove(tmp_path)
        raise


def _write_file(path, fcont, uid=None, gid=None, fsync=True):

    uid = uid or config.uid
    gid = gid or config.gid

    with open(path, 'w') as f:
        f.write(fcont)
        f.flush()
        if fsync:
            os.fsync(f.fileno())

    if uid is not None and gid is not None:
        os.chown(path, uid, gid)


def remove(path, ignore_errors=False, onerror=None):

    if onerror is None:
        onerror = _on_error_raise

    if ignore_errors:
        onerror = _on_error_pass

    try:
        is_dir = os.path.isdir(path)
    except os.error:
        onerror(os.path.isdir, path, sys.exc_info())
        return

    if not is_dir:
        try:
            os.remove(path)
        except os.error:
            onerror(os.remove, path, sys.exc_info())
        return

    names = []
    try:
        names = os.listdir(path)
    except os.error:
        onerror(os.listdir, path, sys.exc_info())

    for name in names:
        fullname = os.path.join(path, name)
        remove(fullname, ignore_errors, onerror)

    try:
        os.rmdir(path)
    except os.error:
        onerror(os.rmdir, path, sys.exc_info())


def _on_error_pass(*args, **kwargs):
    pass


def _on_error_raise(*args, **kwargs):
    raise


def calc_checksums(path, sha1=False, md5=False, crc32=False, sha256=False,
                   block_size=READ_BLOCK, io_limit=READ_BLOCK):

    checksums = {
        'sha1': None,
        'md5': None,
        'crc32': None,
        'sha256': None
    }

    if (sha1 or md5 or crc32 or sha256) is False:
        return checksums

    if block_size <= 0:
        raise FSUtilError('block_size must be positive integer')

    if io_limit == 0:
        raise FSUtilError('io_limit shoud not be zero')

    min_io_time = float(block_size) / io_limit

    sum_sha1 = hashlib.sha1()
    sum_md5 = hashlib.md5()
    sum_crc32 = 0
    sum_sha256 = hashlib.sha256()

    with open(path, 'rb') as f_path:

        while True:
            t0 = time.time()

            buf = f_path.read(block_size)
            if buf == '':
                break

            t1 = time.time()

            time_sleep = max(0, min_io_time - (t1 - t0))
            if time_sleep > 0:
                time.sleep(time_sleep)

            if sha1:
                sum_sha1.update(buf)
            if md5:
                sum_md5.update(buf)
            if crc32:
                sum_crc32 = binascii.crc32(buf, sum_crc32)
            if sha256:
                sum_sha256.update(buf)

    if sha1:
        checksums['sha1'] = sum_sha1.hexdigest()
    if md5:
        checksums['md5'] = sum_md5.hexdigest()
    if crc32:
        checksums['crc32'] = '%08x' % (sum_crc32 & 0xffffffff)
    if sha256:
        checksums['sha256'] = sum_sha256.hexdigest()

    return checksums


def _to_dict(_namedtuple):
    return dict(_namedtuple._asdict())



#!/usr/bin/env python
# coding: utf-8

from pykit import rangeset
from pykit import redisutil
from pykit import utdocker
from pykit import ututil
from pykit import zktx
from pykit import zkutil
from pykit.zktx.status import COMMITTED
from pykit.zktx.test import base

dd = ututil.dd

redis_tag = 'daocloud.io/redis:3.2.3'
redis_port = 6379

zk_tag = 'daocloud.io/zookeeper:3.4.10'
zk_name = 'zk_test'


class TestSlave(base.ZKTestBase):

    @classmethod
    def setUpClass(cls):
        utdocker.pull_image(redis_tag)
        utdocker.pull_image(zk_tag)

    def setUp(self):
        super(TestSlave, self).setUp()
        utdocker.create_network()

        utdocker.start_container('redis-0', redis_tag, '192.168.52.40', '')
        dd('started redis in docker')

        redisutil.wait_serve(('192.168.52.40', redis_port))
        self.redis_cli = redisutil.get_client(('192.168.52.40', redis_port))

        self.txid_path = "tx/txidset"
        self.storage = zktx.RedisStorage(self.redis_cli, self.txid_path)

        self.zk.create('tx', 'tx')
        self.zk.create('tx/txidset', '{}')
        self.zk.create('tx/alive', '{}')
        self.zk.create('tx/txid_maker', '{}')
        self.zk.create('tx/journal', '{}')

        for d in ('lock', 'record'):
            self.zk.create(d, '{}')
            for d1 in ('leader', 'meta', 'job'):
                self.zk.create('/'.join([d, d1]), '{}')
                for d2 in ('block_group', 'global', 'region', 'server'):
                    self.zk.create('/'.join([d, d1, d2]), '{}')

        self.zke, _ = zkutil.kazoo_client_ext(self.zk)

        self.slave = zktx.Slave(self.zke, self.storage)

    def tearDown(self):
        utdocker.remove_container('redis-0')
        utdocker.remove_container(zk_name)

    def _dump_to_zk(self, kv):
        with zktx.ZKTransaction(self.zk) as t1:

            for k, v in kv.items():
                r = t1.lock_get(k)
                r.v = v
                t1.set(r)

            t1.commit()

    def test_apply(self):
        cases = (
            {'meta/server/k1': 10, 'meta/region/k2': 100},
            {'meta/server/k3': 'server_v1', 'meta/region/k4': 'region_v2'},

            {'meta/server/k5': {}, 'meta/region/k6': {}},
            {'meta/server/k7': {'x': 'x', 'y': 'y'}, 'meta/region/k8': {'x': 'x', 'y': 'y'}},

            {'meta/server/k9': [], 'meta/region/k10': []},
            {'meta/server/k11': [1, 2], 'meta/region/k12': [4, 5]},
        )

        max_txid = 1
        for c in cases:
            self._dump_to_zk(c)
            cmt_txidset = self.storage.txidset.get()[COMMITTED]

            if max_txid == 1:
                self.assertEqual([], cmt_txidset)

            self.slave.apply()

            for k, v in c.items():
                k_parts = k.split('/', 2)
                actual_val = self.storage.record.hget(k_parts[1], k_parts[2])
                self.assertEqual(v, actual_val)

            ex = rangeset.RangeSet([[1, max_txid + 1]])
            cmt_txidset = self.storage.txidset.get()[COMMITTED]
            self.assertEqual(ex, cmt_txidset)

            max_txid += 1

        # no update
        self.slave.apply()
        ex = rangeset.RangeSet([[1, max_txid]])
        cmt_txidset = self.storage.txidset.get()[COMMITTED]
        self.assertEqual(ex, cmt_txidset)

    def test_jour_not_found(self):
        c = {
            'meta/server/k1': 1,
            'meta/server/k2': 'server_v1',
            'meta/server/k3': {},
            'meta/server/k4': {'x': 'x'},
            'meta/server/k5': [],
            'meta/server/k6': [1, 'x'],

            'meta/region/k1': 1,
            'meta/region/k2': 'region_v1',
            'meta/region/k3': {},
            'meta/region/k4': {'x': 'x'},
            'meta/region/k5': [],
            'meta/region/k6': [1, 'x'],
        }

        self._dump_to_zk(c)

        self.zk.delete('tx/journal/0000000001')

        self.slave.apply()

        for k, v in c.items():
            k_parts = k.split('/', 2)
            actual_val = self.storage.record.hget(k_parts[1], k_parts[2])
            self.assertEqual(v, actual_val)

        ex = rangeset.RangeSet([[1, 2]])
        cmt_txidset = self.storage.txidset.get()[COMMITTED]
        self.assertEqual(ex, cmt_txidset)

        # apply
        c = {
            'meta/server/k1': 100,
        }
        self._dump_to_zk(c)
        self.slave.apply()

        ex = rangeset.RangeSet([[1, 3]])
        cmt_txidset = self.storage.txidset.get()[COMMITTED]
        self.assertEqual(ex, cmt_txidset)

        actual_val = self.storage.record.hget('server', 'k1')
        self.assertEqual(100, actual_val)



import logging
import os
import time

from pykit import fsutil

logger = logging.getLogger(__name__)


class ParseLogError(Exception):
    pass


def build_entry(context, log_name, file_name, log_str, log_conf):
    log_entry = {
        'log_name': log_name,
        'log_file': file_name,
        'content': log_str,
        'node_id': context['node_id'],
        'node_ip': context['node_ip'],
        'count': 1,
    }

    try:
        log_info = log_conf['parse'](log_str)

    except Exception as e:
        logger.exception('failed to parse log: %s, %s, %s' %
                         (log_name, log_str, repr(e)))
        raise ParseLogError('faild to parse log: %s' % log_name)

    log_entry.update(log_info)

    return log_entry


def put_into_cache(log_cache, log_entry):
    log_ts = log_entry['log_ts']
    source_file = log_entry['source_file']
    line_number = log_entry['line_number']

    if log_ts not in log_cache:
        log_cache[log_ts] = {}

    if source_file not in log_cache[log_ts]:
        log_cache[log_ts][source_file] = {}

    cache_source_file = log_cache[log_ts][source_file]

    if line_number not in cache_source_file:
        cache_source_file[line_number] = log_entry
        return

    old_entry = cache_source_file[line_number]

    log_entry['count'] = old_entry['count'] + 1

    cache_source_file[line_number] = log_entry
    return


def _iter_log(log_conf):
    file_path = log_conf['file_path']
    log_lines = []

    try:
        for line in fsutil.Cat(file_path).iterate(
                timeout=3, default_seek=-1024*1024*2):
            if log_conf['is_first_line'](line):
                if len(log_lines) > 0:
                    yield ''.join(log_lines)
                    log_lines = []

                log_lines = [line]
            else:
                if len(log_lines) < 100:
                    log_lines.append(line)

    except Exception as e:
        logger.info('got exception: %s when iter lines of file: %s' %
                    (repr(e), file_path))

        if len(log_lines) > 0:
            yield ''.join(log_lines)


def iter_log(log_conf):

    while True:
        for log_str in _iter_log(log_conf):
            yield log_str

        time.sleep(1)


def _scan(context, log_name):
    log_stat = context['stat'][log_name]
    log_conf = context['conf'][log_name]
    log_cache = context['cache'][log_name]

    file_path = log_conf['file_path']
    file_name = os.path.basename(file_path)

    log_stat['total_n'] = 0
    log_stat['reported_n'] = 0

    for log_str in iter_log(log_conf):
        log_str = log_str[:1024]
        log_stat['total_n'] += 1

        log_level = log_conf['get_level'](log_str)

        if log_level not in log_conf['level']:
            continue

        log_entry = build_entry(context, log_name, file_name,
                                log_str, log_conf)

        log_stat['latence'] = time.time() - log_entry['log_ts']

        log_stat['reported_n'] += 1

        with context['cache_lock']:
            put_into_cache(log_cache, log_entry)


def scan(context, log_name):
    while True:
        try:
            _scan(context, log_name)

        except Exception as e:
            logger.exception('failed to scan log: %s, %s' %
                             (log_name, repr(e)))

            context['stat'][log_name]['error'] = repr(e)
            time.sleep(1)
